from __future__ import absolute_import

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import calendar

from kivy.properties import ObjectProperty

from es_gui.resources.widgets.common import ResultsViewer


class PerformanceResultsViewer(ResultsViewer):
    """The screen for displaying plots inside the application or exporting results."""

    current_fig = ObjectProperty()
    current_ax = ObjectProperty()

    def __init__(self, **kwargs):
        super(PerformanceResultsViewer, self).__init__(**kwargs)

        self.dfs = {}
        self.variables = []

        self.time_selector.start_time.bind(on_text_validate=self.draw_figure)
        self.time_selector.end_time.bind(on_text_validate=self.draw_figure)

    def on_pre_enter(self):
        """Set toolbar."""
        self.rv.data = self.manager.handler.get_solved_sims()
        self.variables = ['Energy', 'Hourly HVAC Power', 'Hourly Battery Power', 'Efficiency']
        self._update_toolbar()

    def _update_toolbar(self, *args):
        """Update the data viewing toolbar based on selections."""
        super(PerformanceResultsViewer, self)._update_toolbar(self)

        self.vars_button.values = self.variables

    def draw_figure(self, *args):
        """Plot figures."""
        self._update_selection()

        if not self._validate_inputs():
            return

        start_time, end_time = self.time_selector.get_inputs()
        plot_type = self.vars_button.text

        results = self.dfs

        if plot_type in ['total bill', 'total savings']:
            self._reinit_graph(has_legend=False)
        else:
            self._reinit_graph(has_legend=True)

        ax = self.current_ax
        ax.clear()

        plt.xticks(rotation=0)
        plt.grid(True)

        # create dataframes for plotting
        results_ls = [results[key] for key in results]
        resultsF = pd.concat(results_ls, axis=0, ignore_index=True)
        resultsF[['Discharge Power', 'Charge Power']] *= 1000                                   # convert to kW
        resultsF['Battery Power'] = resultsF['Charge Power'] - resultsF['Discharge Power']
        resultsF['Facility Total HVAC Electricity Demand Rate'] /= 1000                         # convert to kW
        resultsF['Discharge Energy'] = resultsF['Discharge Power']*0.25                         # convert to kWh
        resultsF['Charge Energy'] = resultsF['Charge Power']*0.25                               # convert to KWh
        resultsF['HVAC Energy'] = resultsF['Facility Total HVAC Electricity Demand Rate']*0.25  # convert to kWh
        resultsF['Load Percent'] = 100*resultsF['Discharge Energy'].sum()/resultsF['Charge Energy'].sum()
        resultsF['HVAC Percent'] = 100*resultsF['HVAC Energy'].sum()/resultsF['Charge Energy'].sum()
        resultsF['Month'] = [calendar.month_abbr[month_num] for month_num in resultsF['Month']]
        resultsF['Month'] = pd.Categorical(resultsF['Month'], categories=[month_abbr for month_abbr in calendar.month_abbr], ordered=True)
        resultsF['Month'] = resultsF['Month'].cat.remove_unused_categories()
        groupedF = resultsF.groupby(['Month', 'Day', 'Hour'])

        energyF = groupedF.sum().reset_index()
        energyF['Hour of Month'] = (energyF['Day']-1)*24 + energyF['Hour']
        energyF = energyF.loc[energyF['Hour of Month'].between(start_time, end_time, 'both')]

        powerF = groupedF.mean().reset_index()
        powerF['Hour of Month'] = (energyF['Day'] - 1)*24 + energyF['Hour']
        powerF = powerF.loc[powerF['Hour of Month'].between(start_time, end_time, 'both')]

        # make plots
        if plot_type == 'Energy':
            ax = energyF.groupby('Month').sum()[['Discharge Energy', 'Charge Energy', 'HVAC Energy']].plot(kind='bar', ax=ax, legend=True, rot=45)
            ax.set_title('Energy Consumption')
            ax.set_ylabel('kWh')
            ax.legend(bbox_to_anchor=(1, 1), loc='upper left')
        elif plot_type == 'Hourly HVAC Power':
            powerF = powerF.groupby(['Month', 'Day', 'Hour']).mean().unstack(level=0)
            days = powerF.index.get_level_values('Day').unique().to_list()
            max_hours = len(powerF.index)
            ax = powerF['Facility Total HVAC Electricity Demand Rate'].plot(kind='line', ax=ax, legend=True, rot=0, colormap='tab20')
            ax.set_xticks(np.linspace(0, max_hours, len(days), endpoint=False))
            ax.set_xticklabels(days)
            ax.set_ylabel('kW')
            ax.set_xlabel('Day of Month')
            ax.set_title('Hourly HVAC Power Consumption')
            ax.legend(bbox_to_anchor=(1, 1), loc='upper left')
        elif plot_type == 'Hourly Battery Power':
            powerF = powerF.groupby(['Month', 'Day', 'Hour']).mean().unstack(level=0)
            days = powerF.index.get_level_values('Day').unique().to_list()
            max_hours = len(powerF.index)
            ax = powerF['Battery Power'].plot(kind='line', ax=ax, legend=True, rot=0, colormap='tab20')
            ax.set_xticks(np.linspace(0, max_hours, len(days), endpoint=False))
            ax.set_xticklabels(days)
            ax.set_ylabel('kW')
            ax.set_xlabel('Day of Month')
            ax.set_title('Hourly Battery Power Consumption')
            ax.legend(bbox_to_anchor=(1, 1), loc='upper left')
        elif plot_type == 'Efficiency':
            efficiencyF = energyF.groupby('Month').sum()
            ax = (efficiencyF['Discharge Energy']*100/(efficiencyF['Charge Energy']+efficiencyF['HVAC Energy'])).plot(kind='bar', ax=ax, legend=False, rot=45)
            ticks = ax.get_yticks()
            ax.set_yticklabels([str(tick) + '%' for tick in ticks])
            ax.set_yticks(ticks)
            ax.set_title('Monthly Efficiency')

            for container in ax.containers:
                labels = [str(round(num, 1)) + '%' for num in container.datavalues]
                ax.bar_label(container, labels=labels)

        self.plotbox.children[0].draw()

    def _update_selection(self):
        """Update the dictionary of DataFrames whenever new selections of data are made."""
        # Identify the selected run(s) from RunSelector.
        rv = self.rv
        runs_selected = [rv.data[selected_ix] for selected_ix in rv.layout_manager.selected_nodes]

        results = {}

        for run in runs_selected:
            label = run['name']
            df = run['results']

            results[label] = df

        self.dfs = results
        self._update_toolbar()

    def export_png(self):
        """Export currently displayed figure to .png file in specified location."""
        outdir_root = os.path.join('results', 'btm', 'plots')

        super(PerformanceResultsViewer, self).export_png(outdir_root)

    def export_csv(self):
        """Export selected DataFrames to .csv files in specified location."""
        outdir_root = os.path.join('results', 'btm', 'csv')

        super(PerformanceResultsViewer, self).export_csv(outdir_root)



from random import sample, choice
from functools import partial
from collections import OrderedDict
import calendar
import os
import datetime
import webbrowser
import base64
from time import sleep

from jinja2 import Environment, FileSystemLoader, select_autoescape
from kivy.uix.screenmanager import Screen, ScreenManager, WipeTransition, NoTransition
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.modalview import ModalView
from kivy.clock import Clock

from es_gui.proving_grounds.charts import BarChart, StackedBarChart, MultisetBarChart, PieChart, DonutChart
from es_gui.resources.widgets.common import TWO_ABC_WIDTH, THREE_ABC_WIDTH, MyPopup, ReportScreen, PALETTE, WizardReportInterface, ReportChartToggle


class ValuationReport(WizardReportInterface):
    chart_types = OrderedDict({'Revenue (by month)': 'revenue_bar',
                   'Revenue (by stream)': 'revenue_multisetbar',
                   'Participation (total)': 'activity_donut',
                   #'Activity (by source)': 'activity_stackedbar',
                   'Participation (by month)': 'activity_stackedbar_normalized'
                   })

    def __init__(self, chart_data, report_attributes, market=None, **kwargs):
        super(ValuationReport, self).__init__(**kwargs)

        self.chart_type = type
        self.chart_data = chart_data
        self.report_attributes = report_attributes
        self.market = market

        sm = self.report_sm

        # build chart type selection buttons and corresponding ReportScreens
        for opt in self.chart_types.items():
            button = ReportChartToggle(text=opt[0])
            button.bind(state=partial(self.add_report, opt[1]))
            self.chart_type_toggle.add_widget(button)

            screen = ValuationReportScreen(type=opt[1], chart_data=self.chart_data, market=self.market, name=opt[1])
            sm.add_widget(screen)

    def add_report(self, chart_type, *args):
        # adds a ReportScreen of type chart_type to the screen manager
        sm = self.report_sm
        sm.transition = WipeTransition(duration=0.8, clearcolor=[1, 1, 1, 1])

        if not sm.has_screen(chart_type):
            screen = ValuationReportScreen(type=chart_type, chart_data=self.chart_data, market=self.market, name=chart_type)
            sm.add_widget(screen)

        sm.current = chart_type
    
    def open_generate_report_menu(self):
        GenerateReportMenu.host_report = self
        gen_report_menu = GenerateReportMenu()
        gen_report_menu.open()


class ReportScreenManager(ScreenManager):
    transition = WipeTransition(duration=0.8, clearcolor=[1, 1, 1, 1])


class ValuationReportScreen(ReportScreen):
    """"""

    # lookup table for decision variables for each model formulation
    activities = dict()
    activities['arbitrage'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'),]
    activities['pjm_pfp'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'), ('q_reg', 'regulation'), ]
    activities['ercot_arbreg'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'), ('q_ru', 'regulation up'), ('q_rd', 'regulation down'), ]
    activities['miso_pfp'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'), ('q_reg', 'regulation'), ]
    activities['isone_pfp'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'), ('q_reg', 'regulation'), ]
    activities['nyiso_pfp'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'), ('q_reg', 'regulation'), ]
    activities['spp_pfp'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'), ('q_ru', 'regulation up'), ('q_rd', 'regulation down'), ]
    activities['caiso_pfp'] = [('q_r', 'buy (arbitrage)'), ('q_d', 'sell (arbitrage)'), ('q_ru', 'regulation up'), ('q_rd', 'regulation down'), ]

    # lookup table for actions that constitute regulation services for each model formulation
    regulation_def = dict()
    regulation_def['arbitrage'] = []
    regulation_def['pjm_pfp'] = ['regulation', ]
    regulation_def['ercot_arbreg'] = ['regulation up', 'regulation down', ]
    regulation_def['miso_pfp'] = ['regulation', ]
    regulation_def['isone_pfp'] = ['regulation', ]
    regulation_def['nyiso_pfp'] = ['regulation', ]
    regulation_def['spp_pfp'] = ['regulation up', 'regulation down', ]
    regulation_def['caiso_pfp'] = ['regulation up', 'regulation down', ]

    def __init__(self, type, chart_data, market=None, do_animation=True, **kwargs):
        super(ValuationReportScreen, self).__init__(**kwargs)

        self.chart_type = type
        self.chart_data = chart_data
        self.do_animation = do_animation

        # if not market:
        #     # infer market type from Optimizer property
        #     market = self.chart_data[0][1].market_type

        if self.chart_type == 'revenue_bar':
            # bar chart for revenue by month
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = BarChart(bar_spacing=25, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_revenue_bar_chart)
        elif self.chart_type == 'revenue_multisetbar':
            # multiset bar chart for revenue by month by revenue stream
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            self.chart = MultisetBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='${0:,.0f}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=self.generate_revenue_multisetbar_chart)
        elif self.chart_type == 'activity_donut':
            # donut chart for overall device activity
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.66

            self.chart = DonutChart(do_animation=self.do_animation)
            self.bind(on_enter=partial(self.generate_activity_donut_chart, market))
        elif self.chart_type == 'activity_stackedbar':
            # set up vertical orientation for stackedbar chart
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            # generate stackedbar chart
            self.chart = StackedBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='{0:n}', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=partial(self.generate_activity_stackedbar_chart, market, False))
        elif self.chart_type == 'activity_stackedbar_normalized':
            # set up vertical orientation for stackedbar chart
            self.chart_bx.orientation = 'vertical'

            #self.desc.width = THREE_ABC_WIDTH
            self.desc_bx.size_hint_y = 0.33

            # generate stackedbar chart
            self.chart = StackedBarChart(bar_spacing=25, legend_width=TWO_ABC_WIDTH/4, y_axis_format='{0:n}%', size_hint_x=1.0, do_animation=self.do_animation)
            self.bind(on_enter=partial(self.generate_activity_stackedbar_chart, market, True))
        else:
            raise(ValueError('An improper chart type was specified. (got {0})'.format(self.chart_type)))

        self.chart_bx.add_widget(self.chart)

    def rgba_to_fraction(self, rgba):
        """Converts rgb values in int format to fractional values suitable for Kivy."""
        if len(rgba) > 3:
            return float(rgba[0])/255, float(rgba[1])/255, float(rgba[2])/255, rgba[3]
        else:
            return float(rgba[0])/255, float(rgba[1])/255, float(rgba[2])/255, 1

    def on_leave(self):
        # reset the chart
        self.chart.clear_widgets()

    def generate_revenue_bar_chart(self, *args):
        bar_data = []

        # select chart colors
        if len(self.chart_data) > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, len(self.chart_data))

        for ix, op in enumerate(self.chart_data, start=0):
            name = op[0]
            _, _, year, month, _ = name.split(' | ')
            #label = ' '.join([year, month])
            #label = calendar.month_abbr[int(month)]
            label = month

            solved_op = op[1]
            results = solved_op.results

            bar_color = colors[divmod(ix, len(colors))[1]]

            # need convert numpy float to native float for calculations
            bar_data.append([label, self.rgba_to_fraction(bar_color), float(solved_op.gross_revenue)])

        # generate chart
        self.chart.draw_chart(bar_data)
        max_bar = self.chart.max_bar
        min_bar = self.chart.min_bar

        # generate report text
        self.title.text = "Here's how much revenue the device generated each month."
        self.desc.text = 'Revenue was generated based on participation in the selected revenue streams. '
        report_templates = [
            'The gross revenue generated over the evaluation period was [b]${0:,.2f}[/b].'.format(sum([op[1].gross_revenue for op in self.chart_data])),
            'The highest revenue in a month was [b]${0:,.2f}[/b], generated in [b]{1}[/b].'.format(max_bar.value,
                                                                                                   calendar.month_name[list(calendar.month_abbr).index(max_bar.name)]),
        'The lowest revenue in a month was [b]${0:,.2f}[/b], generated in [b]{1}[/b].'.format(min_bar.value,
                                                                                              calendar.month_name[list(calendar.month_abbr).index(min_bar.name)]),
        ]

        self.desc.text += ' '.join(report_templates)

    def generate_revenue_multisetbar_chart(self, *args):
        n_rev_cats = 2

        # select chart colors
        if n_rev_cats > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, n_rev_cats)

        multisetbar_data = OrderedDict()

        # compute activity counts
        for ix, op in enumerate(self.chart_data, start=0):
            name = op[0]
            _, _, year, month, _ = name.split(' | ')
            # label = ' '.join([year, month])
            #label = calendar.month_abbr[int(month)]
            label = month

            solved_op = op[1]
            results = solved_op.results

            try:
                rev_arb = float(results['rev_arb'].tail(1))
                rev_reg = float(results['rev_reg'].tail(1))
            except TypeError:
                rev_arb = 0
                rev_reg = 0

            bar_group = [['arbitrage', self.rgba_to_fraction(colors[0]), rev_arb]]
            bar_group.append(['regulation', self.rgba_to_fraction(colors[1]), rev_reg])

            multisetbar_data[label] = bar_group

        # generate chart
        self.chart.draw_chart(multisetbar_data)

        # generate report text
        self.title.text = "Here's how the device generated revenue each month."
        self.desc.text = 'Revenue was generated based on participation in the selected revenue streams. '
        report_templates = [
            'The [b]gross revenue[/b] generated over the evaluation period was [b]${0:,.2f}[/b].'.format(sum([op[1].gross_revenue for op in self.chart_data])),
        ]

        try:
            total_rev_arb = sum([float(op[1].results['rev_arb'].tail(1)) for op in self.chart_data])
        except TypeError:
            total_rev_arb = 0

        if total_rev_arb >= 0:
            rev_arb_format = '${:,.2f}'.format(total_rev_arb)
        else:
            rev_arb_format = '-${:,.2f}'.format(-total_rev_arb)

        if total_rev_arb < 0:
            report_templates.append('The gross revenue from [b]arbitrage[/b] was [b]{0}[/b], an overall deficit. ' \
                                    'This implies participation in arbitrage was solely for the purpose of having capacity to offer regulation up services.'.format(rev_arb_format))

        self.desc.text += ' '.join(report_templates)

    def generate_activity_donut_chart(self, market=None, *args):
        donut_data = []
        n_activities = dict()
        n_total = 0

        try:
            activities = self.activities[market]
            regulation_def = self.regulation_def[market]
        except KeyError:
            raise(ValueError('Invalid or no market type/formulation specified.'))

        n_activity_cats = len(activities)

        # select chart colors
        if n_activity_cats > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, n_activity_cats)

        # compute activity counts
        for ix, (var, name) in enumerate(activities, start=0):
            n_activity = sum([len(op[1].results[var].values.nonzero()[0]) for op in self.chart_data])
            n_total += n_activity

            n_activities[name] = n_activity

            donut_data.append([name, self.rgba_to_fraction(colors[ix]), n_activity])

        # generate chart
        self.chart.draw_chart(donut_data)

        # generate report text
        self.title.text = "Here's how much the device participated in each revenue stream."
        self.desc.text = 'Each sector represents the share of total state of charge management actions attributed to the specific activity. For example, if 33% of the actions were \"buy (arbitrage),\" that means 33% of the total actions performed were for buying energy through arbitrage. \n\n'
        report_templates = [
            'The [b]total[/b] number of actions performed over the evaluation period was [b]{0:n}[/b].'.format(
                n_total),
        ]

        if n_activities['sell (arbitrage)']/float(n_total) < n_activities['buy (arbitrage)']/float(n_total):
            report_templates.append(
                'The [b]percentage[/b] of actions of selling in the [b]arbitrage[/b] market was [b]{0:.2f}%[/b], less than the percentage for buying. ' \
                'This implies that participation in arbitrage was for the purpose of having energy to offer regulation up services.'.format(
                    n_activities['sell (arbitrage)']/float(n_total)*100))
        #
        # if sum([n_activities[name] for name in regulation_def])/float(n_total) > 0.50:
        #     report_templates.append(
        #         'The device spent a majority of its actions on offering regulation services.'
        #     )

        self.desc.text += ' '.join(report_templates)

    def generate_activity_stackedbar_chart(self, market=None, normalized=False, *args):
        """
        Creates a stacked bar chart counting the number of times each decision variable was nonzero. The number and labels of bar components are determined by the market type/formulation.

        :param market: str for the name of the market type/formulation. This is used for the lookup table in the Screen properties.
        :param normalized: Boolean; if true, bar stacks add up to 100% and each stack element represents percentage of the total.

        :return:
        """
        stackedbar_data = OrderedDict()

        try:
            activities = self.activities[market]
            regulation_def = self.regulation_def[market]
        except KeyError:
            raise(ValueError('Invalid or no market type/formulation specified.'))

        n_activity_cats = len(activities)

        # select chart colors
        if n_activity_cats > len(PALETTE):
            colors = PALETTE
        else:
            colors = sample(PALETTE, n_activity_cats)

        # compute activity counts
        for ix, op in enumerate(self.chart_data, start=0):
            name = op[0]
            _, _, year, month, _ = name.split(' | ')
            # label = ' '.join([year, month])
            #label = calendar.month_abbr[int(month)]
            label = month

            solved_op = op[1]
            results = solved_op.results

            bar_stack = []

            if normalized:
                # use percentages instead of counts for bar value

                # compute the total number of actions in the stack
                n_month_total = float(sum([len(results[var].values.nonzero()[0]) for (var, cat) in activities]))

                for iy, (var, cat) in enumerate(activities, start=0):
                    component_color = colors[iy]
                    n_activity = sum([len(results[var].values.nonzero()[0])])

                    try:
                        n_activity_normalized = n_activity/n_month_total*100
                    except ZeroDivisionError:
                        n_activity_normalized = 0

                    bar_stack.append([cat, self.rgba_to_fraction(component_color), n_activity_normalized])
            else:
                # use counts for bar value
                for iy, (var, cat) in enumerate(activities, start=0):
                    component_color = colors[iy]
                    n_activity = sum([len(results[var].nonzero()[0])])

                    bar_stack.append([cat, self.rgba_to_fraction(component_color), n_activity])

            stackedbar_data[label] = bar_stack

        # generate chart
        self.chart.draw_chart(stackedbar_data)

        # generate report text
        self.title.text = "Here's how the device participated in each revenue stream each month."

        if normalized:
            self.desc.text = 'Each bar segment represents the share of total actions attributed to each activity per month. \n\n'
        else:
            self.desc.text = 'Each bar segment represents the number of times the corresponding activity was performed during that month. \n\n'

        report_templates = [
        ]

        self.desc.text += ' '.join(report_templates)


class GenerateReportMenu(ModalView):
    host_report = None
    graphics_locations = {}
    report_id = None

    def __init__(self, **kwargs):
        super(GenerateReportMenu, self).__init__(**kwargs)

        self.sm.transition = NoTransition()

        # Assign a virtually unique ID to this report generator instance
        self.report_id = '_'.join([
            datetime.datetime.now().strftime('%Y_%m_%d_%H%M%S'), 
            self.host_report.report_attributes['market area']
        ]
        )

    def save_figure(self, screen, *args):
        if not self.sm.has_screen(screen.name):
            self.sm.add_widget(screen)

        self.sm.current = screen.name
        chart_images_dir = os.path.join('results', 'valuation', 'report', self.report_id, 'images')
        os.makedirs(chart_images_dir, exist_ok=True)

        chart_save_location = os.path.join(chart_images_dir, 'chart_{n}.png'.format(n=screen.name))

        Clock.schedule_once(partial(screen.chart.export_to_png, chart_save_location), 1.0)

        # Save image name/path for report generator.
        self.graphics_locations[screen.name] = os.path.join('images', 'chart_{n}.png'.format(n=screen.name))

    def generate_report_screens(self):
        screen_flip_interval = 1.2
        n_charts = len(self.host_report.chart_types.items())

        # Draw figures for saving to .png.
        for ix, opt in enumerate(self.host_report.chart_types.items(), start=0):
            screen = ValuationReportScreen(type=opt[1], chart_data=self.host_report.chart_data, market=self.host_report.market, name=opt[1], do_animation=False)

            Clock.schedule_once(partial(self.save_figure, screen), ix * screen_flip_interval)

        self.generate_report_button.disabled = True

        # Generate report.
        Clock.schedule_once(lambda dt: self.generate_report_from_template(), (n_charts+1)*screen_flip_interval)

    def generate_report_from_template(self):
        # Get current date.
        now = datetime.datetime.now()
        today = now.strftime("%B %d, %Y")

        # Get report-specific data.
        report_attributes = self.host_report.report_attributes

        ISO = report_attributes['market area']
        pricing_node = report_attributes['pricing node']
        ES_device = report_attributes['selected device']
        revenue_streams = report_attributes['revenue streams']
        dates_analyzed = report_attributes['dates analyzed']
        power_rating = report_attributes['Power_rating']
        energy_capacity = report_attributes['Energy_capacity']
        storage_efficiency = report_attributes['Self_discharge_efficiency']
        conversion_efficiency = report_attributes['Round_trip_efficiency']

        # Retrieve HTML template based on selected ISO.
        template_dir = os.path.join('es_gui', 'resources', 'report_templates')
        output_dir_name = self.report_id

        output_dir = os.path.join('results', 'valuation', 'report', output_dir_name)
        os.makedirs(output_dir, exist_ok=True)

        # Initialize Jinja environment.
        env = Environment(loader=FileSystemLoader(template_dir), autoescape=select_autoescape(['html']))
        fname = os.path.join(output_dir, '{0}_valuation_report.html'.format('_'.join([ISO, pricing_node])))

        if ISO == "ERCOT":
            template = env.get_template('valuation_report_ERCOT.html')
            
        elif ISO == "MISO":
            template = env.get_template('valuation_report_MISO.html')
            
        elif ISO == "PJM":
            template = env.get_template('valuation_report_PJM.html')

        elif ISO == "ISONE":
            template = env.get_template('valuation_report_ISONE.html')
        elif ISO == "NYISO":
            template = env.get_template('valuation_report_NYISO.html')
        elif ISO == "SPP":
            template = env.get_template('valuation_report_SPP.html')
        elif ISO == "CAISO":
            template = env.get_template('valuation_report_CAISO.html')
        else :
            raise ValueError('The selected ISO does not have a reporting template.')

        # Render output file.
        output = template.render(
					 # GENERAL OUTPUT
					 today=today,
			  		 market_area=ISO,
			  		 header="This report shows the results from optimizations performed by QuESt Valuation.",
			  		 pricing_node=pricing_node,						 
			  		 dates_analyzed=dates_analyzed,
			  		 revenue_streams=revenue_streams,
			  		 ES_device=ES_device,
                     QuESt_Logo=os.path.join('..', 'images', 'static', 'Quest_Logo_RGB.png'),
			  		 SNL_image=os.path.join('..', 'images', 'static', 'SNL.png'),
			  		 DOE_image=os.path.join('..', 'images', 'static', 'DOE.png'),
			  		 acknowledgement="Sandia National Laboratories is a multimission laboratory managed and operated by National Technology & Engineering Solutions of Sandia, LLC, a wholly owned subsidiary of Honeywell International Inc., for the U.S. Department of Energy's National Nuclear Security Administration under contract DE-NA0003525.",
					 # PARAMETER VALUES TABLE
					 power_rating=power_rating,
					 energy_capacity=energy_capacity,
					 storage_efficiency=storage_efficiency,
					 conversion_efficiency=conversion_efficiency,						 
					 # FIGURES
					 revenue_total=self.graphics_locations['revenue_bar'],
					 revenue_source=self.graphics_locations['revenue_multisetbar'],
					 activity_total_percent=self.graphics_locations['activity_donut'],
					 activity_source=self.graphics_locations['activity_stackedbar_normalized'],
					 activity_source_percent=self.graphics_locations['activity_stackedbar_normalized']
					 )

        with open(fname,"w") as f:
            f.write(output)
        
        completion_popup = OpenGeneratedReportPopup()
        completion_popup.report_filename = fname
        completion_popup.open()


class OpenGeneratedReportPopup(MyPopup):
    report_filename = ''

    def open_generated_report(self):
        """Opens the generated report and dismisses the popup,"""
        webbrowser.open('file://' + os.path.realpath(self.report_filename))
        self.dismiss()


"""Example for running valuation simulations."""
from __future__ import absolute_import

import logging
import matplotlib.pyplot as plt

from op_handler import ValuationOptimizerHandler as val_op


if __name__ == "__main__":
    with open("valuation_optimizer.log", "w"):
        pass

    logging.basicConfig(
        filename="valuation_optimizer.log",
        format="[%(levelname)s] %(asctime)s: %(message)s",
        level=logging.INFO,
    )

    params = [
        {
            "Power_rating": 2,  # MW
            "Energy_capacity": 4,  # MWh
            "Self_discharge_efficiency": 1,
            "R": 0,  # Interest rate [hour^-1]
            "Round_trip_efficiency": 0.85,
            "Reserve_reg_min": 0,  # Fraction of q_reg bid to increase state of charge minimum by.
            "Reserve_reg_max": 1,  # Fraction of q_reg bid to decrease state of charge maximum by.
            "State_of_charge_min": 0,  # Fraction of energy capacity representing the minimum state of charge.
            "State_of_charge_max": 1,  # Fraction of energy capacity representing the maximum state of charge.
            "State_of_charge_init": 0.5,  # Initial state of charge as a fraction of energy capacity.
        }
    ]

    requests_PJM = {
        "iso": "PJM",
        "node id": 1,  # PJM-rTO
        "market type": ["arbitrage", "pfp"],
    }

    requests_MISO = {
        "iso": "MISO",
        "node id": "MICHIGAN.HUB",
        "market type": ["arbitrage", "pfp"],
    }

    requests_ISONE = {
        "iso": "ISONE",
        "node id": 4004,  # CONNECTICUT HUB
        "market type": ["arbitrage", "pfp"],
    }

    requests_NYISO = {
        "iso": "NYISO",
        "node id": 61761,  # NYC ZONE
        "market type": ["arbitrage", "pfp"],
    }

    requests_CAISO = {
        "iso": "CAISO",
        "node id": "TH_SP15_GEN-APND",  # SOUTH HUB
        "market type": ["arbitrage", "pfp"],
    }

    requests_SPP = {
        "iso": "SPP",
        "node id": "AECI",
        "market type": ["arbitrage", "pfp"],
    }

    requests_ERCOT = {
        "iso": "ERCOT",
        "node id": "HB_HOUSTON",
        "market type": ["arbitrage", "pfp"],
    }

    months = [(1, 2018), (2, 2018), (3, 2018)]
    requests = [
        requests_PJM,
        requests_MISO,
        requests_ISONE,
        requests_NYISO,
        requests_CAISO,
        requests_SPP,
        requests_ERCOT,
    ]
    handler = val_op(
        solver_name="glpk",
        data_path="apps/valuation/data/",
        save_name="valuation_dms.p",
    )

    for request in requests:
        request["param set"] = params
        request["months"] = months
        market_types = request["market type"]
        for market_type in market_types:
            request["market type"] = market_type

            handler.process_requests(request)

    solved_ops = handler.get_solved_ops()

    results_dict = {
        "PJM": [],
        "MISO": [],
        "ISONE": [],
        "NYISO": [],
        "CAISO": [],
        "SPP": [],
        "ERCOT": [],
    }

    for result in solved_ops:
        if result["iso"] == "PJM":
            results_dict["PJM"].append(result)
        elif result["iso"] == "MISO":
            results_dict["MISO"].append(result)
        elif result["iso"] == "ISONE":
            results_dict["ISONE"].append(result)
        elif result["iso"] == "NYISO":
            results_dict["NYISO"].append(result)
        elif result["iso"] == "CAISO":
            results_dict["CAISO"].append(result)
        elif result["iso"] == "SPP":
            results_dict["SPP"].append(result)
        elif result["iso"] == "ERCOT":
            results_dict["ERCOT"].append(result)

    fig_ls = []
    for iso in results_dict:
        fig, ax = plt.subplots()
        arb = []
        pfp = []
        for result in results_dict[iso]:
            if result["market type"] == "arbitrage":
                arb.append((result["month"], result["optimizer"].gross_revenue))
            else:
                pfp.append((result["month"], result["optimizer"].gross_revenue))

        ax.bar(
            x=[date[0] for date in arb],
            height=[date[1] for date in arb],
            width=-0.5,
            align="edge",
            color=["blue"],
        )
        ax.bar(
            x=[date[0] for date in pfp],
            height=[date[1] for date in pfp],
            width=0.5,
            align="edge",
            color=["orange"],
        )
        ax.legend()
        ax.set_title("{} Gross Revenue 2018".format(iso))
        ax.set_ylabel("$")
        fig_ls.append((fig, ax))
        fig_ls.append((fig, ax))
        fig_ls.append((fig, ax))



import os
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

color_red, color_blue = np.array([247, 140, 102])/255, np.array([102, 163, 224])/255
cBlue = np.array([0, 102, 204])/255
cAmber = np.array([255, 136, 0])/255
cGreen = np.array([108, 179, 18])/255
cOrange = np.array([242, 63, 0])/255
cBlueGrey = np.array([125, 142, 160])/255

plt.style.use(os.path.join('es_gui', 'apps', 'tech_selection', 'mpl_style_presentations.mplstyle'))


def plot_table_feasibility(df_data, figsize=(7, 5),
                           cmap=[color_red, color_blue],
                           xlabel=None, ylabel=None, xticklabels=[], yticklabels=[],
                           update_individual_xticklabel=None):
    """"Display feasibility tables as heatmap plots."""
    
    plt.style.use(os.path.join('es_gui', 'apps', 'tech_selection', 'mpl_style_presentations.mplstyle'))

    # Default labels for the x and y ticklabels, if none are given
    if len(xticklabels) == 0: xticklabels = df_data.columns
    if len(yticklabels) == 0: yticklabels = df_data.index

    # Create feasibility table (as a heatmap plot)
    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(df_data, cmap=cmap, cbar=False, linewidths=0.1,
                annot=df_data.applymap(lambda x: 'Yes' if x == 1 else 'No'),
                fmt='s', annot_kws={'fontsize': 2, 'fontfamily': 'sans-serif', 'va': 'center_baseline'},
                xticklabels=xticklabels, yticklabels=yticklabels, ax=ax)
    ax.tick_params(rotation=0)

    # Update settings for x and y labels
    ax.set_xlabel(xlabel, fontsize=12, weight='bold')
    ax.set_ylabel(ylabel, fontsize=12, weight='bold')

    # Update settings for ticks and their labels
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position('top')
    ax.tick_params(bottom=False, top=False, left=False, right=False)

    # Update settings for individual x ticklabels
    if update_individual_xticklabel is not None:
        for xtick_idx, xtick_props in update_individual_xticklabel.items():
            ax.get_xticklabels()[xtick_idx].update(xtick_props)
            
    # Return figure handle
    return fig


def plot_ranking_techs(df_data):
    fig, axs = plt.subplots(ncols=5, sharey=True, figsize=(1.9, 1))
    for column, color, ax in zip(['Application score', 'Location score', 'Cost score', 'Maturity score', 'Total score'],
                                 [cBlue, cAmber, cGreen, cOrange, cBlueGrey], axs):
        df_data[column].plot.barh(ax=ax, fc=color)
        ax.set_xticks([0, 0.5, 1])
        ax.set_xticklabels(['0', '0.5', '1'])
        ax.set_xlim(0, 1)
        ax.set_ylabel('')
        ax.set_title(column.replace(' score', ''))

    return fig


from __future__ import absolute_import

import os
from functools import partial

from kivy.uix.screenmanager import Screen
from kivy.app import App

from es_gui.tools.valuation.valuation_dms import ValuationDMS
from es_gui.resources.widgets.common import WarningPopup, NavigationButton
from es_gui.proving_grounds.help_carousel import HelpCarouselModalView
from .op_handler import ValuationOptimizerHandler


class ValuationHomeScreen(Screen):
    """
    The home screen for doing energy storage valuation analysis.
    """
    def __init__(self, **kwargs):
        super(ValuationHomeScreen, self).__init__(**kwargs)

        # Initialize data management system.
        self.dms = ValuationDMS(max_memory=App.get_running_app().config.getint('valuation', 'valuation_dms_size')*1000,
                                save_data=bool(App.get_running_app().config.getint('valuation', 'valuation_dms_save')),
                                save_name='valuation_dms.p',
                                home_path='data')
        self.handler = ValuationOptimizerHandler(App.get_running_app().config.get('optimization', 'solver'))
        self.handler.dms = self.dms

    def on_enter(self):
        ab = self.manager.nav_bar
        ab.reset_nav_bar()
        ab.set_title('Valuation')

        help_button = NavigationButton(
            text='help',
            on_release=self.open_help_carousel,
        )

        ab.action_view.add_widget(help_button)

        # data_manager = App.get_running_app().data_manager
        
        # # Check if any data is available.
        # if not data_manager.data_bank:
        #     no_data_popup = WarningPopup()
        #     no_data_popup.popup_text.text = "Looks like you haven't downloaded any data yet. Try using QuESt Data Manager to get some data before returning here!"
        #     no_data_popup.dismiss_button.text = "Got it, take me back!"

        #     no_data_popup.bind(on_dismiss=partial(ab.go_to_screen, 'index'))
        #     no_data_popup.open()
    
    def open_help_carousel(self, *args):
        """
        """
        help_carousel_view = HelpCarouselModalView()
        help_carousel_view.title.text = "QuESt Valuation"

        slide_01_text = "QuESt Valuation is an application for estimating the revenue potential for an energy storage system providing ISO/RTO services. It takes a retrospective analysis approach using historical data.\n\nData is required to use the tools in QuESt Valuation. Options in the user interface such as market area selection are entirely based on the contents of your QuESt data bank."

        slide_02_text = "The Wizard mode under the Simulation tab walks you through a series of prompts to set up the analysis. This mode is streamlined for a simpler experience compared to the Batch Runs mode.\n\nYou will need the following data to use this tool:\n* Market data for each ISO/RTO that you want to look at"

        slide_03_text = "Upon completion of the wizard, you will be taken to the summary report screen. There a number of reports you can browse through that summarize different aspects of the simulation results. A brief synopsis of each component of the results including some key numbers.\n\nThe 'Generate report' button can be used to produce a document that summarizes the wizard run."

        slide_04_text = "This document includes your input selections, a primer on the mathematical model used, and all of the charts from the wizard summary reports.\n\nThe resulting HTML document and images are saved to the /results/*/report directory. You can view the report in a web browser."

        slide_05_text = "The Batch Runs mode is more advanced than the Wizard mode. The workflow for using this mode is identical but offers more flexibility and options.\n\nThere are two panels for your input: 'Data' and 'Parameters'. You can toggle between the two using the buttons at the bottom. Note that changing market area or pricing node selections will reset the other input widgets."

        slide_06_text = "The parameters available are more numerous than those in the Wizard mode and may also vary among the market areas. The default values indicated by the hint text for each parameter will be used if you do not enter in a new value.\n\nTry using the 'Tab' key to quickly navigate among the text input fields."

        slide_07_text = "The parameter sweep can be used for parameter sensitivity analysis. The sweep will be applied to each month of data selected.\n\nIn this example, a sweep over the power rating from 5 to 20 MW in ten evenly-spaced points will be performed for each month selected in the 'Data' tab. If four months were selected then a total of 40 models will be solved."

        slide_08_text = "When you are done with your selections, click the 'Go!' button to proceed. The models will be built and solved in the background. Upon completion, a popup will open to inform you of the results and provide any pertinent warnings.\n\nYou can proceed to the QuESt Valuation results viewer via the popup, navigation bar, or QuESt Valuation home screen to look at the results."

        slide_09_text = "You can view simulation results in more detail using the Results Viewer tool. You can plot time series data and export simulation results for external processing."

        slides = [
            (os.path.join("es_gui", "resources", "help_views", "valuation", "01.png"), slide_01_text),
            (os.path.join("es_gui", "resources", "help_views", "valuation", "02.png"), slide_02_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "wizard_report", "01.png"), slide_03_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "wizard_report", "02.png"), slide_04_text),
            (os.path.join("es_gui", "resources", "help_views", "valuation", "03.png"), slide_05_text),
            (os.path.join("es_gui", "resources", "help_views", "valuation", "04.png"), slide_06_text),
            (os.path.join("es_gui", "resources", "help_views", "valuation", "05.png"), slide_07_text),
            (os.path.join("es_gui", "resources", "help_views", "valuation", "06.png"), slide_08_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "results_viewer", "00.png"), slide_09_text),
        ]

        help_carousel_view.add_slides(slides)
        help_carousel_view.open()



"""Data Manager for Valuation Tool."""

from __future__ import absolute_import
import logging
import os

from quest_eras.quest_library.utilities.pomo.dms import DataManagementSystem
import quest_eras.quest_library.utilities.pomo.valuation.pjm.data_processing_tools as dpt


class ValuationDMSPJM(DataManagementSystem):
    """
    A class for managing data for the energy storage valuation optimization functions.

    Class methods for each type of file to be loaded are included, extending from the get_data() method of the superclass.
    Each of these methods uses get_data() to retrieve the relevant data and loads the file and adds it to the DMS if the data is not loaded.
    An optional class method for calling each of the individual data methods can be included to form the necessary arguments and return the desired variables.

    :param home_path: A string indicating the relative path to where data is saved.
    """

    def __init__(self, home_path, **kwargs):
        DataManagementSystem.__init__(self, **kwargs)

        self.home_path = home_path

        # with open(os.path.abspath(os.path.join(self.home_path, '..', 'es_gui', 'apps', 'valuation', 'definitions', 'nodes.json')), 'r') as fp:
        #     self.NODES = json.load(fp)

        # self.node_names = pd.read_excel(self.home_path+'nodeid.xlsx', sheetname=None)
        self.delimiter = " @ "  # delimiter used to split information in id_key

    def get_node_name(self, node_id, ISO):
        """
        Retrieve the node name corresponding to the given node_id using the lookup table loaded during initialization.

        :param node_id: A string or int of a node ID.
        :return: The corresponding node name as a string.
        """
        # TODO: map node_id to node name

        return str(node_id)

    def get_pjm_lmp_data(self, *args):
        """Retrieve LMP data for PJM (deprecated since 1.0)."""
        logging.info("DMS: Loading PJM DA-LMP")
        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(*args)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da = dpt.read_pjm_da_lmp(*args)
            self.add_data(lmp_da, *args)
        finally:
            return lmp_da

    def get_pjm_reg_price_data(self, *args):
        """Retrieve regulation price data for PJM (deprecated since 1.0)."""
        logging.info("DMS: Loading PJM regulation prices")
        try:
            # attempt to access data if it is already loaded
            RegCCP = self.get_data(*args + ("RegCCP",))
            RegPCP = self.get_data(*args + ("RegPCP",))
        except KeyError:
            # load the data and add it to the DMS
            RegCCP, RegPCP = dpt.read_pjm_reg_price(*args)
            self.add_data({"RegCCP": RegCCP, "RegPCP": RegPCP}, *args)
        finally:
            return RegCCP, RegPCP

    def get_pjm_mileage_data(self, *args):
        """Retrieve regulation mileage data for PJM (deprecated since 1.0)."""
        logging.info("DMS: Loading PJM mileage data")
        try:
            # attempt to access data if it is already loaded
            MR = self.get_data(*args + ("MR",))
            RA = self.get_data(*args + ("RA",))
            RD = self.get_data(*args + ("RD",))
        except KeyError:
            # load the data and add it to the DMS
            MR, RA, RD = dpt.read_pjm_mileage(*args)
            self.add_data({"MR": MR, "RA": RA, "RD": RD}, *args)
        finally:
            return MR, RA, RD

    def get_pjm_reg_signal_data(self, *args):
        """Retrieve regulation signal data for PJM (deprecated since 1.0)."""
        logging.info("DMS: Loading PJM regulation signal")
        try:
            # attempt to access data if it is already loaded
            RUP = self.get_data(*args + ("RegUp",))
            RDW = self.get_data(*args + ("RegDown",))
        except KeyError:
            # load the data and add it to the DMS
            RUP, RDW = dpt.read_pjm_reg_signal(*args)
            self.add_data({"RegUp": RUP, "RegDown": RDW}, *args)
        finally:
            return RUP, RDW

    def get_pjm_data(self, year, month, nodeid):
        """Retrieve data for PJM."""
        path = os.path.join(self.home_path, "PJM")

        nodeid = str(nodeid)
        year = str(year)
        month = str(month)

        lmp_key = self.delimiter.join([path, year, month, nodeid, "SPP"])
        mr_key = self.delimiter.join([path, year, month, "MR"])
        ra_key = self.delimiter.join([path, year, month, "RA"])
        rd_key = self.delimiter.join([path, year, month, "RD"])
        rccp_key = self.delimiter.join([path, year, month, "RegCCP"])
        rpcp_key = self.delimiter.join([path, year, month, "RegPCP"])

        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(lmp_key)
            MR = self.get_data(mr_key)
            RA = self.get_data(ra_key)
            RD = self.get_data(rd_key)
            RegCCP = self.get_data(rccp_key)
            RegPCP = self.get_data(rpcp_key)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da, MR, RA, RD, RegCCP, RegPCP = dpt.read_pjm_data(
                path, year, month, nodeid
            )

            self.add_data(lmp_da, lmp_key)
            self.add_data(MR, mr_key)
            self.add_data(RA, ra_key)
            self.add_data(RD, rd_key)
            self.add_data(RegCCP, rccp_key)
            self.add_data(RegPCP, rpcp_key)

        return lmp_da, MR, RA, RD, RegCCP, RegPCP


# if __name__ == "__main__":
#     dms = ValuationDMS(save_name="valuation_dms.p", home_path="data")

#     # # ERCOT - data doesn't exist
#     # year = 2010
#     # month = 1
#     # settlement_point = 'HB_HOUSTON'

#     # spp_da, rd, ru = dms.get_ercot_data(year, month, settlement_point)

#     # # ERCOT
#     # year = 2010
#     # month = 12
#     # settlement_point = 'HB_HOUSTON'

#     # spp_da, rd, ru = dms.get_ercot_data(year, month, settlement_point)

#     # PJM
#     # year = 2016
#     # month = 5
#     # nodeid = 1

#     # lmp_da, MR, RA, RD, RegCCP, RegPCP = dms.get_pjm_data(year, month, nodeid)

#     # MISO
#     year = 2015
#     month = 3
#     nodeid = "AEC"

#     lmp_da, RegMCP = dms.get_miso_data(year, month, nodeid)



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 17 08:21:28 2021.

@author: wolis
"""

from __future__ import absolute_import

import logging
import calendar
import pandas as pd
from eppy.modeleditor import IDF

from kivy.clock import Clock

from es_gui.tools.performance.Battery_v2 import Battery as bt
from es_gui.tools.performance.Grid_simulator import Grid_simulator as gs
from es_gui.resources.widgets.common import WarningPopup


class PerformanceSimHandler:
    """A handler for creating and solving Performance instances as requested."""

    dms = None
    solved_sims = []

    # day names, variables, actuators, and idd for EnergyPlus
    days = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
    variables = [("Zone Mean Air Temperature", "Zone One"), ('Site Outdoor Air Drybulb Temperature', 'Environment'),
                 ("Zone Thermostat Heating Setpoint Temperature", "Zone One"),
                 ("Zone Thermostat Cooling Setpoint Temperature", "Zone One"), ('Facility Total Building Electricity Demand Rate', 'Whole Building'),
                 ('Facility Total HVAC Electricity Demand Rate', 'Whole Building'), ('Facility Total Electricity Demand Rate', 'Whole Building'),
                 ('Electric Equipment Electricity Rate', 'BATTERY'), ('Site Total Zone Exfiltration Heat Loss', 'Environment'),
                 ('Site Total Zone Exhaust Air Heat Loss', 'Environment'), ('Zone Exfiltration Heat Transfer Rate', 'Zone One'),
                 ('Zone Exhaust Air Heat Transfer Rate', 'Zone One'), ('Zone Air Heat Balance Internal Convective Heat Gain Rate', 'Zone One'),
                 ('Zone Air Heat Balance Surface Convection Rate', 'Zone One'), ('Zone Air Heat Balance Outdoor Air Transfer Rate', 'Zone One'),
                 ('Zone Air Heat Balance System Air Transfer Rate', 'Zone One'), ('Zone Air Heat Balance System Convective Heat Gain Rate', 'Zone One'),
                 ('Zone Packaged Terminal Air Conditioner Total Heating Rate',
                  'Zone One PTAC'), ('Zone Packaged Terminal Air Conditioner Total Cooling Rate', 'Zone One PTAC'),
                 ('Zone Packaged Terminal Air Conditioner Electricity Rate', 'Zone One PTAC')]
    actuators = [('Zone Temperature Control', 'Heating Setpoint', 'Zone One'),
                 ('Zone Temperature Control', 'Cooling Setpoint', 'Zone One'),
                 ('ElectricEquipment', 'Electricity Rate', 'BATTERY')]
    idd = 'energyplus/Energy+.idd'
    IDF.setiddname(idd)

    def __init__(self, output_dir):
        self._output_dir = output_dir

        self.bad_run_popup = BadRunPopup()

    @property
    def output_dir(self):
        """Directory for EnergyPlus output files."""
        return self._output_dir

    @output_dir.setter
    def output_dir(self, value):
        self._output_dir = value

    def process_requests(self, requests, *args):
        """Generate and solve EnergyPlus models based on the given requests."""
        output_dir = self.output_dir
        variables = self.variables
        actuators = self.actuators

        # get rv data
        hvac_name = requests['hvac']['name']
        hvac_path = requests['hvac']['path']
        location_name = requests['location']['name']
        location_path = requests['location']['path']
        params = requests['params']

        # reset EnergyPlus input file with selected information
        idf_obj = IDF(hvac_path)
        runperiods = idf_obj.idfobjects['RunPeriod']
        while len(runperiods) > 1:
            runperiods.pop(-1)

        if len(requests['profile']) == 1 and not requests['profile'][0]['path'] == 1:
            profile_name = requests['profile'][0]['name']
            profile_path = requests['profile'][0]['path']

            cdF = pd.read_excel(profile_path)
            cdF[['P_d', 'P_c']] /= 1000

            idf_obj = IDF(hvac_path)
            battery_obj = idf_obj.idfobjects['ElectricEquipment'][0]
            battery_obj.Design_Level = params['pRat']*1e6
            runperiod_obj = runperiods[0]
            runperiod_obj.Begin_Month = cdF['Month'].iloc[0]
            runperiod_obj.End_Month = cdF['Month'].iloc[-1]
            runperiod_obj.Begin_Year = cdF['Year'].iloc[0]
            runperiod_obj.End_Year = cdF['Year'].iloc[-1]
            month_range = calendar.monthrange(cdF['Year'].iloc[0], cdF['Month'].iloc[0])
            runperiod_obj.Begin_Day_of_Month = 1
            runperiod_obj.End_Day_of_Month = month_range[1]
            runperiod_obj.Day_of_Week_for_Start_Day = calendar.day_name[month_range[0]]

        else:
            cdF = pd.DataFrame(columns=['P_c', 'P_d', 'Month'])

            for i, profile in enumerate(requests['profile']):
                profile_op = profile['op']
                try:
                    resultsF, revenue = profile_op['optimizer'].get_results()
                except ValueError:
                    resultsF = profile_op['optimizer'].get_results()

                try:
                    month_num = list(calendar.month_abbr).index(profile_op['month'])
                except ValueError:
                    month_num = list(calendar.month_name).index(profile_op['month'])

                try:
                    year = profile_op['year']
                except KeyError:
                    year = 2020

                try:
                    resultsF['setpoint'] = resultsF['q_r'] + resultsF['q_rd'] - resultsF['q_reg'] - resultsF['q_d'] - resultsF['q_ru']
                except KeyError:
                    resultsF['P_c'] = resultsF['Pcharge']/1000  # convert to MW for E+
                    resultsF['P_d'] = resultsF['Pdischarge']/1000  # convert to MW for E+
                else:
                    resultsF['P_c'] = [value if value >= 0 else 0 for value in resultsF['setpoint']]
                    resultsF['P_d'] = [-value if value <= 0 else 0 for value in resultsF['setpoint']]

                cd = pd.concat([resultsF.reset_index()]*4).sort_values('index', axis=0).reset_index(drop=True)[['P_c', 'P_d']]
                cd['Month'] = [month_num]*len(cd.index)
                cdF = cdF.append(cd)

                try:
                    runperiod_obj = runperiods[i]
                    runperiod_obj.Name = 'Run Period {}'.format(i)
                    runperiod_obj.Begin_Month = runperiod_obj.End_Month = int(month_num)
                    runperiod_obj.Begin_Year = runperiod_obj.End_Year = int(year)
                    month_range = calendar.monthrange(int(year), int(month_num))
                    runperiod_obj.Begin_Day_of_Month = 1
                    runperiod_obj.End_Day_of_Month = month_range[1]
                    runperiod_obj.Day_of_Week_for_Start_Day = calendar.day_name[month_range[0]]
                except IndexError:
                    idf_obj.copyidfobject(runperiods[0])
                    runperiod_obj = idf_obj.idfobjects['RunPeriod'][-1]
                    runperiod_obj.Name = 'Run Period {}'.format(i)
                    runperiod_obj.Begin_Month = runperiod_obj.End_Month = int(month_num)
                    runperiod_obj.Begin_Year = runperiod_obj.End_Year = int(year)
                    month_range = calendar.monthrange(int(year), int(month_num))
                    runperiod_obj.Begin_Day_of_Month = 1
                    runperiod_obj.End_Day_of_Month = month_range[1]
                    runperiod_obj.Day_of_Week_for_Start_Day = calendar.day_name[month_range[0]]
                else:
                    print('Successful runperiod change.')

                if month_num == 5:
                    runperiod_obj.Begin_Month = 4
                    runperiod_obj.Begin_Day_of_Month = 30

            try:
                battery_obj = idf_obj.idfobjects['ElectricEquipment'][0]
                assert battery_obj.Name == 'BATTERY', 'Battery not first ElectricEquipment object...creating new'
                battery_obj.Design_Level = params['pRat']*1e6
                idf_obj.idfobjects['ElectricEquipment'][0] = battery_obj
            except IndexError or AssertionError:
                idf_obj.newidfobject("Schedule:Constant")
                new_schedule = idf_obj.idfobjects['Schedule:Constant'][-1]
                new_schedule.Name = "Battery Heat Loss Schedule"
                new_schedule.Schedule_Type_Limits_Name = "Fraction"
                new_schedule.Hourly_Value = 0

                idf_obj.newidfobject("ElectricEquipment")
                battery_obj = idf_obj.idfobjects['ElectricEquipment'][-1]
                battery_obj.Name = "BATTERY"
                battery_obj.Zone_or_ZoneList_Name = idf_obj.idfobjects['Zone'][-1].Name
                battery_obj.Schedule_Name = "Battery Heat Loss Schedule"
                battery_obj.Design_Level_Calculation_Method = "EquipmentLevel"
                battery_obj.Design_Level = params['pRat']*1e6
                battery_obj.Watts_per_Zone_Floor_Area = ''
                battery_obj.Watts_per_Person = ''
                battery_obj.Fraction_Latent = 0
                battery_obj.Fraction_Radiant = 0
                battery_obj.Fraction_Lost = 0

                idf_obj.idfobjects["Schedule:Constant"][-1] = new_schedule
                idf_obj.idfobjects["ElectricEquipment"][-1] = battery_obj
                print("New battery and schedule object added to idf...")
            else:
                print("Successful battery update")

        index = None
        for i, obj in enumerate(idf_obj.idfobjects['Construction']):
            if obj['Name'] == 'Outer Shell':
                index = i

        idf_obj.popidfobject('Construction', index)
        if params['insulation'] == 0:
            idf_obj.newidfobject('Construction', Name='Outer Shell', Outside_Layer='Corten Steel')
        else:
            ins_mat = idf_obj.idfobjects['Material:NoMass'][-1]
            ins_mat.Thermal_Resistance = params['insulation']
            idf_obj.newidfobject('Construction', Name='Outer Shell', Outside_Layer='Corten Steel', Layer_2='Insulation')

        idf_obj.idfobjects['RunPeriod'] = runperiods
        idf_obj.save()

        # run EnergyPlus simulation
        battery = bt(params['eCap'], params['pRat'], params['n_s'], params['n_p'],
                     params['q_rate'], params['v_rate'], params['r'], params['k'])
        g_sim = gs(hvac_path, location_path, output_dir, variables=variables, actuators=actuators,
                   battery=battery)
        g_sim.h_setpoint = params['h_setpoint']
        g_sim.c_setpoint = params['c_setpoint']
        g_sim.power_electronics_flag = False
        g_sim.hvac_flag = False
        g_sim.charge_discharge = cdF
        return_code = g_sim.new_simulation()

        # Energyplus has crashed...restart QuESt
        if not return_code == 0:
            self.bad_run_popup.bad_open()

        results = g_sim.get_results()

        # save results
        if len(requests['profile']) == 1:
            self._save_to_solved_ops(results, hvac_name, location_name, profile_name)
        else:
            for profile in requests['profile']:
                profile_op = profile['op']
                try:
                    month_num = list(calendar.month_abbr).index(profile_op['month'])
                except ValueError:
                    month_num = list(calendar.month_name).index(profile_op['month'])

                self._save_to_solved_ops(results.loc[results['Month'] == month_num], hvac_name, location_name, profile['name'])

        logging.info('Op Handler: Finished processing requested jobs.')

        return results

    def _solve_model(self, op):
        op.solver = self.solver_name
        op.run()

        return op

    def _save_to_solved_ops(self, results, hvac, location, profile):
        """Save the results of simulations."""
        name = ' | '.join([hvac, location, profile])

        results_dict = {}
        results_dict['name'] = name
        results_dict['results'] = results
        results_dict['hvac'] = hvac
        results_dict['location'] = location
        results_dict['profile'] = profile

        self.solved_sims.append(results_dict)

        return (name, results)

    def get_solved_sims(self):
        """Return the list of solved  objects in reverse chronological order."""
        return self.solved_sims


class BadRunPopup(WarningPopup):
    """Popup for EnergyPlus crash."""

    def __init__(self, **kwargs):
        super(BadRunPopup, self).__init__(**kwargs)

        self.popup_text.text = 'EnergyPlus had a fatal error. Please read the log or restart the QuESt App to continue.'

    def bad_open(self):
        """Open popup from main thread."""
        Clock.schedule_once(self.open)



from __future__ import absolute_import, print_function

import logging
from functools import partial
import webbrowser
import calendar
import os
import numpy as np
import threading
import json

from kivy.uix.screenmanager import Screen, ScreenManager, SlideTransition, ScreenManagerException
from kivy.uix.label import Label
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import ObjectProperty, StringProperty, BooleanProperty, ListProperty, DictProperty
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.animation import Animation
from kivy.clock import Clock, mainthread
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.app import App
from kivy.uix.popup import Popup
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.image import Image
from kivy.uix.modalview import ModalView
from kivy.uix.textinput import TextInput

# from es_gui.apps.valuation.reporting import Report
from .reporting import BtmCostSavingsReport
from es_gui.resources.widgets.common import BodyTextBase, MyPopup, WarningPopup, TileButton, RecycleViewRow, InputError, BASE_TRANSITION_DUR, BUTTON_FLASH_DUR, ANIM_STAGGER, FADEIN_DUR, SLIDER_DUR, PALETTE, rgba_to_fraction, fade_in_animation, slow_blinking_animation, WizardCompletePopup, ParameterRow, ParameterGridWidget
from es_gui.proving_grounds.data_importer import DataImporter
from es_gui.apps.data_manager.data_manager import DATA_HOME
from es_gui.tools.btm.readutdata import get_pv_profile_string


class CostSavingsWizard(Screen):
    """The main screen for the cost savings wizard. This hosts the nested screen manager for the actual wizard."""
    def on_enter(self):
        ab = self.manager.nav_bar
        ab.set_title('Time-of-Use Cost Savings')

        # self.sm.generate_start()

    def on_leave(self):
        # Reset wizard to initial state by removing all screens except the first.
        self.sm.current = 'start'

        if len(self.sm.screens) > 1:
            self.sm.clear_widgets(screens=self.sm.screens[1:])


class CostSavingsWizardScreenManager(ScreenManager):
    """The screen manager for the cost savings wizard screens."""
    def __init__(self, **kwargs):
        super(CostSavingsWizardScreenManager, self).__init__(**kwargs)

        self.transition = SlideTransition()
        self.add_widget(CostSavingsWizardStart(name='start'))


class CostSavingsWizardStart(Screen):
    """The starting/welcome screen for the cost savings wizard."""
    def _next_screen(self):
        if not self.manager.has_screen('rate_select'):
            screen = CostSavingsWizardRateSelect(name='rate_select')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'rate_select'


class CostSavingsWizardRateSelect(Screen):
    """The starting/welcome screen for the cost savings wizard."""
    rate_structure_selected = DictProperty()
    has_selection = BooleanProperty(False)

    def __init__(self, **kwargs):
        super(CostSavingsWizardRateSelect, self).__init__(**kwargs)

        CostSavingsRateStructureRVEntry.host_screen = self
    
    def on_enter(self):
        try:
            data_manager = App.get_running_app().data_manager
            rate_structure_options = [rs[1] for rs in data_manager.get_rate_structures().items()]
            self.rate_structure_rv.data = rate_structure_options
            self.rate_structure_rv.unfiltered_data = rate_structure_options
        except KeyError as e:
            logging.warning('CostSavings: No rate structures available to select.')
            # TODO: Warning popup
        
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)
    
    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0
    
    def on_rate_structure_selected(self, instance, value):
        try:
            logging.info('CostSavings: Rate structure selection changed to {0}.'.format(value['name']))
        except KeyError:
            logging.info('CostSavings: Rate structure selection reset.')
            self.rate_structure_desc.text = ''
            self.has_selection = False
        else:
            Animation.stop_all(self.preview_box, 'opacity')
            self.preview_box.opacity = 0

            def _generate_preview():
                self.generate_schedule_charts()
                self.generate_flat_demand_rate_table()
                self.generate_misc_table()

                Clock.schedule_once(partial(fade_in_animation, self.preview_box), 0)
                self.has_selection = True
            
            thread_preview = threading.Thread(target=_generate_preview)
            thread_preview.start()

    def on_has_selection(self, instance, value):
        self.next_button.disabled = not value            

    def generate_flat_demand_rate_table(self):
        """Generates the preview table for the flat demand rate structure."""
        flat_demand_rates = self.rate_structure_selected['demand rate structure'].get('flat rates', {})

        table_data = [str(flat_demand_rates.get(month, '')) for month in calendar.month_abbr[1:]]
        self.flat_demand_rates_table.populate_table(table_data)
        
    def generate_misc_table(self):
        """Generates the preview table for the miscellaneous information."""
        self.misc_data_table.populate_table(self.rate_structure_selected)

    def generate_schedule_charts(self, *args):
        """Generates the preview for the weekday and weekend rate schedule charts."""
        weekday_schedule_data = self.rate_structure_selected['energy rate structure'].get('weekday schedule', [])
        weekend_schedule_data = self.rate_structure_selected['energy rate structure'].get('weekend schedule', [])

        legend_labels = ['${0}/kWh'.format(v) for _,v in self.rate_structure_selected['energy rate structure'].get('energy rates').items()]

        if weekday_schedule_data and weekend_schedule_data:
            n_tiers = len(np.unique(weekday_schedule_data))

            palette = [rgba_to_fraction(color) for color in PALETTE][:n_tiers]
            labels = calendar.month_abbr[1:]

            self.energy_weekday_chart.draw_chart(np.array(weekday_schedule_data), palette, labels, legend_labels=legend_labels)
            self.energy_weekend_chart.draw_chart(np.array(weekend_schedule_data), palette, labels, legend_labels=legend_labels)
        
        weekday_schedule_data = self.rate_structure_selected['demand rate structure'].get('weekday schedule', [])
        weekend_schedule_data = self.rate_structure_selected['demand rate structure'].get('weekend schedule', [])

        legend_labels = ['${0}/kW'.format(v) for _,v in self.rate_structure_selected['demand rate structure'].get('time of use rates').items()]

        if weekday_schedule_data and weekend_schedule_data:
            n_tiers = len(np.unique(weekday_schedule_data))

            # Select chart colors.
            palette = [rgba_to_fraction(color) for color in PALETTE][:n_tiers]
            labels = calendar.month_abbr[1:]

            # Draw charts.
            self.demand_weekday_chart.draw_chart(np.array(weekday_schedule_data), palette, labels, legend_labels=legend_labels)
            self.demand_weekend_chart.draw_chart(np.array(weekend_schedule_data), palette, labels, legend_labels=legend_labels)
    
    def _validate_inputs(self):
        # TODO: Progress already impeded until a structure is selected so...
        return self.rate_structure_selected
    
    def get_selections(self):
        """Retrieves screen's selections."""
        try:
            rate_structure_selected = self._validate_inputs()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            return rate_structure_selected

    def _next_screen(self):
        if not self.manager.has_screen('load_profile_select'):
            screen = CostSavingsWizardLoadSelect(name='load_profile_select')
            self.manager.add_widget(screen)
        
        try:
            self.get_selections()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            self.manager.transition.duration = BASE_TRANSITION_DUR
            self.manager.transition.direction = 'left'
            self.manager.current = 'load_profile_select'


class CostSavingsRateStructureRVEntry(RecycleViewRow):
    host_screen = None

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        super(CostSavingsRateStructureRVEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.rate_structure_selected = rv.data[self.index]


class FlatDemandRateTable(GridLayout):
    """A preview table to show the monthly flat demand rate structure."""
    def reset_table(self):
        """Builds the table column headers if necessary and removes all data entries."""
        if not self.col_headers.children:
            for month in calendar.month_abbr[1:]:
                col_header = BodyTextBase(text=month, color=rgba_to_fraction(PALETTE[1]), font_size=20)
                self.col_headers.add_widget(col_header)
        
        while len(self.data_grid.children) > 0:
            for widget in self.data_grid.children:
                self.data_grid.remove_widget(widget)
    
    def populate_table(self, data):
        """Populates the data entries."""
        self.reset_table()

        for datum in data:
            rate_label = BodyTextBase(text=datum, font_size=20)
            self.data_grid.add_widget(rate_label)


class MiscRateStructureDataTable(GridLayout):
    """A preview table for miscellaneous rate structure information."""
    def reset_table(self):
        """Removes all data entries."""
        while len(self.peak_min_box.children) > 0:
            for widget in self.peak_min_box.children:
                self.peak_min_box.remove_widget(widget)

        while len(self.peak_max_box.children) > 0:
            for widget in self.peak_max_box.children:
                self.peak_max_box.remove_widget(widget)
                
        while len(self.net_metering_type_box.children) > 0:
            for widget in self.net_metering_type_box.children:
                self.net_metering_type_box.remove_widget(widget)
        
        while len(self.energy_sell_price_box.children) > 0:
            for widget in self.energy_sell_price_box.children:
                self.energy_sell_price_box.remove_widget(widget)
    
    def populate_table(self, rate_structure_dict):
        """Populates the table with data from the rate structure dictionary."""
        self.reset_table()

        net_metering_dict = rate_structure_dict['net metering']
        demand_structure_dict = rate_structure_dict['demand rate structure']

        peak_kw_min = str(demand_structure_dict.get('minimum peak demand', '0'))
        peak_kw_min_label = TextInput(text=peak_kw_min, font_size=20, readonly=True)
        self.peak_min_box.add_widget(peak_kw_min_label)

        peak_kw_max = str(demand_structure_dict.get('maximum peak demand', 'None'))
        peak_kw_max_label = TextInput(text=peak_kw_max, font_size=20, readonly=True)
        self.peak_max_box.add_widget(peak_kw_max_label)

        net_metering_type = '2.0' if net_metering_dict['type'] else '1.0'
        net_metering_type_label = TextInput(text=net_metering_type, font_size=20, readonly=True)
        self.net_metering_type_box.add_widget(net_metering_type_label)

        energy_sell_price = 'N/A' if not net_metering_dict.get('energy sell price', False) else str(net_metering_dict.get('energy sell price', ''))
        energy_sell_price_label = TextInput(text=energy_sell_price, font_size=20, readonly=True)
        self.energy_sell_price_box.add_widget(energy_sell_price_label)
        

class CostSavingsWizardLoadSelect(Screen):
    """The load profile selection screen for the cost savings wizard."""
    load_profile_selected = DictProperty()
    has_selection = BooleanProperty(False)
    imported_data_selected = BooleanProperty(False)

    def __init__(self, **kwargs):
        super(CostSavingsWizardLoadSelect, self).__init__(**kwargs)

        LoadProfileRecycleViewRow.load_profile_screen = self

    def on_enter(self):
        try:
            data_manager = App.get_running_app().data_manager

            load_profile_options = [{'name': x[0], 'path': x[1]} for x in data_manager.get_load_profiles().items()]
            self.load_profile_rv.data = load_profile_options
            self.load_profile_rv.unfiltered_data = load_profile_options
        except KeyError as e:
            logging.warning('CostSavings: No load profiles available to select.')
        
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)
    
    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0
    
    def on_load_profile_selected(self, instance, value):
        try:
            logging.info('CostSavings: Load profile selection changed to {0}.'.format(value['name']))
        except KeyError:
            logging.info('CostSavings: Load profile selection reset.')
            self.has_selection = False
        else:
            self.has_selection = True
            self.imported_data_selected = False

    def on_has_selection(self, instance, value):
        self.next_button.disabled = not value
    
    def on_imported_data_selected(self, instance, value):
        if value:
            self.open_data_importer_button.text = 'Data imported'
            self.open_data_importer_button.background_color = rgba_to_fraction(PALETTE[3])
            Clock.schedule_once(partial(slow_blinking_animation, self.open_data_importer_button), 0)

            self.load_profile_rv.deselect_all_nodes()
        else:
            self.open_data_importer_button.text = 'Open data importer'
            self.open_data_importer_button.background_color = rgba_to_fraction(PALETTE[0])
            self.open_data_importer_button.opacity = 1
            Animation.cancel_all(self.open_data_importer_button, 'opacity')
    
    def open_data_importer(self):
        write_directory = os.path.join(DATA_HOME, 'load', 'imported')
        self.data_importer = DataImporter(
            write_directory=write_directory, 
            format_description="The data units should be in kilowatts and there should be 8,760 samples (hourly for one standard year). The time series is assumed to run January through December at an hourly resolution."
        )
        self.data_importer.title.text = "Import a load time series"

        def _check_data_importer_on_dismissal():
            try:
                import_filename = self.data_importer.get_import_selections()
            except (ValueError, AttributeError):
                logging.warning('DataImporter: Nothing was imported.')
            except KeyError:
                logging.warning('DataImporter: Import process was terminated early.')
            else:
                logging.info('DataImporter: Data import complete.')
                self.load_profile_selected = {'name': 'Custom', 'path': import_filename}
                self.imported_data_selected = True
        
        self.data_importer.bind(on_dismiss=lambda t: _check_data_importer_on_dismissal())
        self.data_importer.open()
    
    def _validate_inputs(self):
        # TODO: Progress already impeded until a profile is selected so...
        return self.load_profile_selected
    
    def get_selections(self):
        """Retrieves screen's selections."""
        try:
            load_profile_selected = self._validate_inputs()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            return load_profile_selected
    
    def _next_screen(self):
        if not self.manager.has_screen('pv_profile_select'):
            screen = CostSavingsWizardPVSelect(name='pv_profile_select')
            self.manager.add_widget(screen)
        
        try:
            self.get_selections()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            self.manager.transition.duration = BASE_TRANSITION_DUR
            self.manager.transition.direction = 'left'
            self.manager.current = 'pv_profile_select'


class LoadProfileRecycleViewRow(RecycleViewRow):
    """The representation widget for node in the load profile selector RecycleView."""
    load_profile_screen = None

    def on_touch_down(self, touch):
        """Add selection on touch down."""
        if super(LoadProfileRecycleViewRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        self.selected = is_selected

        if is_selected:
            self.load_profile_screen.load_profile_selected = rv.data[self.index]


class CostSavingsWizardPVSelect(Screen):
    """The optional PV profile selection screen for the cost savings wizard."""
    pv_profile_selected = DictProperty()
    has_selection = BooleanProperty(False)
    imported_data_selected = BooleanProperty(False)

    def __init__(self, **kwargs):
        super(CostSavingsWizardPVSelect, self).__init__(**kwargs)

        PVProfileRecycleViewRow.pv_profile_screen = self

    def on_enter(self):
        try:
            data_manager = App.get_running_app().data_manager

            pv_profile_options = [
                {
                'name': x[0], 'path': x[1], 'descriptors': get_pv_profile_string(x[1])
                } 
                for x in data_manager.get_pv_profiles().items()
                ]
            self.pv_profile_rv.data = pv_profile_options
            self.pv_profile_rv.unfiltered_data = pv_profile_options
        except Exception as e:
            print(e)
        
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)
        
    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0
    
    def on_pv_profile_selected(self, instance, value):
        try:
            logging.info('CostSavings: PV profile selection changed to {0}.'.format(value['name']))
        except KeyError:
            logging.info('CostSavings: PV profile selection reset.')
            self.has_selection = False
        else:
            self.has_selection = True
            self.imported_data_selected = False
    
    def on_imported_data_selected(self, instance, value):
        if value:
            self.open_data_importer_button.text = 'Data imported'
            self.open_data_importer_button.background_color = rgba_to_fraction(PALETTE[3])
            Clock.schedule_once(partial(slow_blinking_animation, self.open_data_importer_button), 0)

            self.pv_profile_rv.deselect_all_nodes()
        else:
            self.open_data_importer_button.text = 'Open data importer'
            self.open_data_importer_button.background_color = rgba_to_fraction(PALETTE[0])
            self.open_data_importer_button.opacity = 1
            Animation.cancel_all(self.open_data_importer_button, 'opacity')
    
    def open_data_importer(self):
        write_directory = os.path.join(DATA_HOME, 'pv')

        def _write_pv_profile_json(fname, dataframe):
            """Writes a generic time series dataframe to a PV profile json.

            Parameters
            ----------
            fname : str
                Name of the file to be saved without an extension
            dataframe : Pandas DataFrame
                DataFrame with one Series which is the PV power profile in watts
            
            Returns
            -------
            str
                The save destination of the resulting file.
            """
            pv_profile_template_file = os.path.join('es_gui', 'resources', 'import_templates', 'pv_profile.json')

            with open(pv_profile_template_file, 'r') as f:
                pv_profile_template = json.load(f)
            
            ac_output_w = dataframe.iloc[:, 0].tolist()
            pv_profile_template['outputs']['ac'] = ac_output_w

            save_destination = os.path.join(write_directory, fname + '.json')
            
            with open(save_destination, 'w') as f:
                json.dump(pv_profile_template, f)
            
            return save_destination

        self.data_importer = DataImporter(
            write_directory=write_directory, 
            write_function=_write_pv_profile_json, 
            format_description="The data units should be in watts and there should be 8,760 samples (hourly for one standard year). The time series is assumed to run January through December at an hourly resolution."
            )
        self.data_importer.title.text = "Import a PV power time series"

        def _check_data_importer_on_dismissal():
            try:
                import_filename = self.data_importer.get_import_selections()
            except (ValueError, AttributeError):
                logging.warning('DataImporter: Nothing was imported.')
            except KeyError:
                logging.warning('DataImporter: Import process was terminated early.')
            else:
                logging.info('DataImporter: Data import complete.')
                self.pv_profile_selected = {'name': 'Custom', 'path': import_filename, 'descriptors': get_pv_profile_string(import_filename)}
                self.imported_data_selected = True
        
        self.data_importer.bind(on_dismiss=lambda t: _check_data_importer_on_dismissal())
        self.data_importer.open()
    
    def _validate_inputs(self):
        return self.pv_profile_selected
    
    def get_selections(self):
        """Retrieves screen's selections."""
        try:
            pv_profile_selected = self._validate_inputs()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            return pv_profile_selected
    
    def _next_screen(self):
        if not self.manager.has_screen('system_parameters'):
            screen = CostSavingsWizardSystemParameters(name='system_parameters')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'system_parameters'


class PVProfileRecycleViewRow(RecycleViewRow):
    """The representation widget for node in the PV profile selector RecycleView."""
    pv_profile_screen = None

    def on_touch_down(self, touch):
        """Add selection on touch down."""
        if super(PVProfileRecycleViewRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        self.selected = is_selected

        if is_selected:
            self.pv_profile_screen.pv_profile_selected = rv.data[self.index]
    
    def deselect_node(self):
        super(PVProfileRecycleViewRow, self).deselect_node(self)


class CostSavingsWizardSystemParameters(Screen):
    """"""
    def on_pre_enter(self):
        if not self.param_widget.children:
            self.param_widget.build()
        # while len(self.param_widget.children) > 0:
        #     for widget in self.param_widget.children:
        #         if isinstance(widget, CostSavingsParameterRow):
        #             self.param_widget.remove_widget(widget)
    
    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)
    
    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0
    
    def _validate_inputs(self):
        params = self.param_widget.get_inputs()
        params_desc = self.param_widget.get_input_strings()

        return params, params_desc
    
    def get_selections(self):
        """Retrieves screen's selections."""
        try:
            params, params_desc = self._validate_inputs()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            return params, params_desc
    
    def _next_screen(self):
        if not self.manager.has_screen('summary'):
            screen = CostSavingsWizardSummary(name='summary')
            self.manager.add_widget(screen)
        
        try:
            self.get_selections()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            self.manager.transition.duration = BASE_TRANSITION_DUR
            self.manager.transition.direction = 'left'
            self.manager.current = 'summary'


class CostSavingsParameterWidget(ParameterGridWidget):
    """Grid layout containing rows of parameter adjustment widgets."""
    def build(self):
        # Build the widget by creating a row for each parameter.
        data_manager = App.get_running_app().data_manager
        MODEL_PARAMS = data_manager.get_btm_cost_savings_model_params()

        for param in MODEL_PARAMS:
            row = ParameterRow(desc=param)
            self.add_widget(row)
            setattr(self, param['attr name'], row)


class CostSavingsWizardSummary(Screen):
    """"""
    def get_selections(self):
        sm = self.manager

        op_handler_requests = {}

        op_handler_requests['rate_structure'] = sm.get_screen('rate_select').get_selections()
        op_handler_requests['load_profile'] = sm.get_screen('load_profile_select').get_selections()
        op_handler_requests['pv_profile'] = sm.get_screen('pv_profile_select').get_selections()
        op_handler_requests['params'], op_handler_requests['param desc'] = sm.get_screen('system_parameters').get_selections()

        return op_handler_requests

    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

        op_handler_requests = self.get_selections()

        load_profile = op_handler_requests['load_profile']
        pv_profile = op_handler_requests['pv_profile']
        system_params = op_handler_requests['param desc']
        rate_structure = op_handler_requests['rate_structure']

        # Rate structure label.
        rate_structure_text = '\n'.join([
            '[b]Rate Structure:[/b]',
            rate_structure['name'],
            rate_structure['utility']['utility name'],
            rate_structure['utility']['rate structure name'],
            ])
        self.rate_structure_label.text = rate_structure_text

        # Load profile label.
        load_profile_text = '\n'.join([
            '[b]Load Profile:[/b]',
            load_profile['name']
            ])
        self.load_profile_label.text = load_profile_text

        # PV profile label.
        pv_profile_text = '[b]PV Profile:[/b]\n'
        pv_profile_text += '\n'.join(pv_profile.get('descriptors', ['None selected']))
        self.pv_profile_label.text = pv_profile_text

        # System parameters label.
        system_parameters_text = '[b]System Parameters:[/b]\n'
        system_parameters_text += '\n'.join(system_params)
        self.system_parameters_label.text = system_parameters_text
    
    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0

    def _next_screen(self):
        if not self.manager.has_screen('execute'):
            screen = CostSavingsWizardExecute(name='execute')
            self.manager.add_widget(screen)
        
        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'execute'


class CostSavingsWizardExecute(Screen):
    """The screen for executing the prescribed optimizations for the cost savings wizard."""
    solved_ops = []
    report_attributes = {}

    def on_enter(self):
        self.execute_run()

    # def on_leave(self):
    #     self.progress_label.text = 'This may take a while. Please wait patiently!'

    def execute_run(self):
        btm_home = self.manager.parent.parent.manager.get_screen('btm_home')
        ss = self.manager.get_screen('summary')

        op_handler_requests = ss.get_selections()

        # Send requests to handler.
        handler = btm_home.handler
        handler.solver_name = App.get_running_app().config.get('optimization', 'solver')
        self.solved_ops, handler_status = handler.process_requests(op_handler_requests)

        popup = WizardCompletePopup()

        # Check BtmOp handler status.
        if len(handler_status) > 0:
            if self.solved_ops:
                # At least one model solved successfully.
                popup.title = "Success!*"
                popup.popup_text.text = '\n'.join([
                    'All finished, but we found these issues:',
                ]
                + list(handler_status)
                )
            else:
                # No models solved successfully.
                popup.title = "Hmm..."
                popup.popup_text.text = '\n'.join([
                    'Unfortunately, none of the models were able to be solved. We found these issues:',
                ]
                + list(handler_status)
                )

                popup.results_button.text = "Take me back"
                popup.bind(on_dismiss=lambda x: self.manager.parent.parent.manager.nav_bar.go_up_screen())  # Go back to BTM Home
                popup.open()
                return
        
        self.report_attributes = op_handler_requests

        popup = WizardCompletePopup()

        # if not handler_status:
        #     popup.title = "Success!*"
        #     popup.popup_text.text = "All calculations finished. Press 'OK' to proceed to the results.\n\n*At least one model (month) had issues being built and/or solved. Any such model will be omitted from the results."

        popup.bind(on_dismiss=self._next_screen)
        popup.open()

    def _next_screen(self, *args):
        """Adds the report screen if it does not exist and changes screens to it."""
        report = BtmCostSavingsReport(name='report', chart_data=self.solved_ops, report_attributes=self.report_attributes)
        self.manager.switch_to(report, direction='left', duration=BASE_TRANSITION_DUR)



from __future__ import division, print_function, absolute_import

import logging
import os

from pyomo.environ import *
import pandas as pd
import numpy as np

from es_gui.tools import optimizer
from es_gui.tools.valuation.constraints import ExpressionsBlock


class ValuationOptimizer(optimizer.Optimizer):
    """A framework wrapper class for creating Pyomo ConcreteModels for energy storage valuation."""

    def __init__(self, price_electricity=None,
                 price_reg_up=None, price_reg_down=None,
                 price_reg_serv_up=None, price_reg_serv_down=None,
                 price_regulation=None, price_reg_service=None,
                 cost_charge=None, cost_discharge=None,
                 mileage_mult=None, mileage_mult_ru=None, mileage_mult_rd=None,
                 perf_score=None, perf_score_ru=None, perf_score_rd=None,
                 fraction_reg_up=None, fraction_reg_down=None,
                 market_type='arbitrage',
                 solver='glpk'):

        # TODO: deprecate Perf_score and mileage_ratio

        self._model = ConcreteModel()
        self._market_type = market_type
        self._solver = solver

        self._expressions_block = None

        self._price_electricity = price_electricity

        self._price_regulation = price_regulation
        self._price_reg_up = price_reg_up
        self._price_reg_down = price_reg_down

        self._price_reg_service = price_reg_service
        self._price_reg_serv_up = price_reg_serv_up
        self._price_reg_serv_down = price_reg_serv_down

        self._cost_charge = cost_charge
        self._cost_discharge = cost_discharge

        self._mileage_mult = mileage_mult
        self._mileage_mult_ru = mileage_mult_ru
        self._mileage_mult_rd = mileage_mult_rd

        self._perf_score = perf_score
        self._perf_score_ru = perf_score_ru
        self._perf_score_rd = perf_score_rd

        self._fraction_reg_up = fraction_reg_up
        self._fraction_reg_down = fraction_reg_down

        self._results = None
        self._gross_revenue = None

    @property
    def price_electricity(self):
        """The price for buying electricity via energy arbitrage [$/MWh]."""
        return self._price_electricity

    @price_electricity.setter
    def price_electricity(self, value):
        self._price_electricity = value

    @property
    def price_regulation(self):
        """The price for providing capacity for regulation services in pay-for-performance markets [$/MWh]."""
        return self._price_regulation

    @price_regulation.setter
    def price_regulation(self, value):
        self._price_regulation = value

    @property
    def price_reg_up(self):
        """The price for providing regulation up products [$/MWh]."""
        return self._price_reg_up

    @price_reg_up.setter
    def price_reg_up(self, value):
        self._price_reg_up = value

    @property
    def price_reg_down(self):
        """The price for providing regulation down products [$/MWh]."""
        return self._price_reg_down

    @price_reg_down.setter
    def price_reg_down(self, value):
        self._price_reg_down = value

    @property
    def price_reg_service(self):
        """The price for providing regulation services; based on mileage or performance [$/MWh]."""
        return self._price_reg_service

    @price_reg_service.setter
    def price_reg_service(self, value):
        self._price_reg_service = value

    @property
    def price_reg_serv_up(self):
        """The price for providing regulation up service; based on mileage or performance [$/MWh]."""
        return self._price_reg_serv_up

    @price_reg_serv_up.setter
    def price_reg_serv_up(self, value):
        self._price_reg_serv_up = value

    @property
    def price_reg_serv_down(self):
        """The price for providing regulation down service; based on mileage or performance [$/MWh]."""
        return self._price_reg_serv_down

    @price_reg_serv_down.setter
    def price_reg_serv_down(self, value):
        self._price_reg_serv_down = value

    @property
    def cost_charge(self):
        """The cost of charging the energy storage device [$/MWh]."""
        return self._cost_charge

    @cost_charge.setter
    def cost_charge(self, value):
        self._cost_charge = value

    @property
    def cost_discharge(self):
        """The cost of discharging the energy storage device [$/MWh]."""
        return self._cost_discharge

    @cost_discharge.setter
    def cost_discharge(self, value):
        self._cost_discharge = value

    @property
    def mileage_mult(self):
        """Mileage multiplier or ratio according to the selected market."""
        return self._mileage_mult

    @mileage_mult.setter
    def mileage_mult(self, value):
        self._mileage_mult = value

    @property
    def mileage_mult_ru(self):
        """Mileage multiplier or ratio for the regulation up service according to the selected market."""
        return self._mileage_mult_ru

    @mileage_mult_ru.setter
    def mileage_mult_ru(self, value):
        self._mileage_mult_ru = value

    @property
    def mileage_mult_rd(self):
        """Mileage multiplier or ratio for the regulation down service according to the selected market."""
        return self._mileage_mult_rd

    @mileage_mult_rd.setter
    def mileage_mult_rd(self, value):
        self._mileage_mult_rd = value

    @property
    def perf_score(self):
        """Performance score of the regulation service provided."""
        return self._perf_score

    @perf_score.setter
    def perf_score(self, value):
        self._perf_score = value

    @property
    def perf_score_ru(self):
        """Performance score of the regulation up service provided."""
        return self._perf_score_ru

    @perf_score_ru.setter
    def perf_score_ru(self, value):
        self._perf_score_ru = value

    @property
    def perf_score_rd(self):
        """Performance score of the regulation down service provided."""
        return self._perf_score_rd

    @perf_score_rd.setter
    def perf_score_rd(self, value):
        self._perf_score_rd = value

    @property
    def fraction_reg_up(self):
        """The fraction of regulation up reserve capacity actually employed."""
        return self._fraction_reg_up

    @fraction_reg_up.setter
    def fraction_reg_up(self, value):
        self._fraction_reg_up = value

    @property
    def fraction_reg_down(self):
        """The fraction of regulation down reserve capacity actually employed."""
        return self._fraction_reg_down

    @fraction_reg_down.setter
    def fraction_reg_down(self, value):
        self._fraction_reg_down = value

    @property
    def solver(self):
        """The name of the solver for Pyomo to use, defaults to 'glpk'."""
        return self._solver

    @solver.setter
    def solver(self, value):
        self._solver = value

    @property
    def expressions_block(self):
        """ExpressionsBlock object for setting model objectives and constraints."""
        return self._expressions_block

    @expressions_block.setter
    def expressions_block(self, value):
        self._expressions_block = value

    @property
    def market_type(self):
        """The name of the market formulation to be modeled, defaults to 'arbitrage'."""
        return self._market_type

    @market_type.setter
    def market_type(self, value):
        if isinstance(value, str):
            self._market_type = value
        else:
            raise TypeError('market_type property must be of type str.')

    @property
    def results(self):
        """Pandas DataFrame containing results."""
        return self._results

    @results.setter
    def results(self, value):
        if isinstance(value, pd.DataFrame):
            self._results = value
        else:
            raise TypeError('results must be a Pandas DataFrame.')

    @property
    def gross_revenue(self):
        """The net revenue generated over the time period as solved for in the optimization."""
        return self._gross_revenue

    @gross_revenue.setter
    def gross_revenue(self, value):
        self._gross_revenue = value

    def _set_model_param(self):
        """Sets the model params for the Pyomo ConcreteModel."""
        m = self.model

        # Check if params common to all formulations are set.
        if not hasattr(m, 'Power_rating'):
            # Power rating; equivalently, the maximum energy charged in one hour [MW].
            logging.debug('ValuationOptimizer: No Power_rating provided, setting default...')
            m.Power_rating = 20

        if not hasattr(m, 'R'):
            # Discount/interest rate [hour^(-1)].
            logging.debug('ValuationOptimizer: No R provided, setting default...')
            m.R = 0

        if not hasattr(m, 'Energy_capacity'):
            # Energy capacity [MWh].
            logging.debug('ValuationOptimizer: No Energy_capacity provided, setting default...')
            m.Energy_capacity = 5

        if not hasattr(m, 'Self_discharge_efficiency'):
            # Fraction of energy maintained over one time period.
            logging.debug('ValuationOptimizer: No Self_discharge_efficiency provided, setting default...')
            m.Self_discharge_efficiency = 1.00
        elif getattr(m, 'Self_discharge_efficiency') > 1.0:
            logging.warning('ValuationOptimizer: Self_discharge_efficiency provided is greater than 1.0, interpreting as percentage...')
            m.Self_discharge_efficiency = m.Self_discharge_efficiency/100

        if not hasattr(m, 'Round_trip_efficiency'):
            # Fraction of input energy that gets stored over one time period.
            logging.debug('ValuationOptimizer: No Round_trip_efficiency provided, setting default...')
            m.Round_trip_efficiency = 0.85
        elif getattr(m, 'Round_trip_efficiency') > 1.0:
            logging.warning('ValuationOptimizer: Round_trip_efficiency provided is greater than 1.0, interpreting as percentage...')
            m.Round_trip_efficiency = m.Round_trip_efficiency/100

        if not hasattr(m, 'Reserve_reg_min'):
            # Fraction of q_reg bid to increase state of charge minimum by.
            logging.debug('ValuationOptimizer: No Reserve_reg_min provided, setting default...')
            m.Reserve_reg_min = 0
        elif getattr(m, 'Reserve_reg_min') > 1.0:
            logging.warning('ValuationOptimizer: Reserve_reg_min provided is greater than 1.0, interpreting as percentage...')
            m.Reserve_reg_min = m.Reserve_reg_min/100

        if not hasattr(m, 'Reserve_reg_max'):
            # Fraction of q_reg bid to decrease state of charge maximum by.
            logging.debug('ValuationOptimizer: No Reserve_reg_max provided, setting default...')
            m.Reserve_reg_max = 0
        elif getattr(m, 'Reserve_reg_max') > 1.0:
            logging.warning('ValuationOptimizer: Reserve_reg_max provided is greater than 1.0, interpreting as percentage...')
            m.Reserve_reg_max = m.Reserve_reg_max/100

        if not hasattr(m, 'State_of_charge_min'):
            # Fraction of energy capacity representing the minimum state of charge.
            logging.debug('ValuationOptimizer: No State_of_charge_min provided, setting default...')
            m.State_of_charge_min = 0
        elif getattr(m, 'State_of_charge_min') > 1.0:
            logging.warning('ValuationOptimizer: State_of_charge_min provided is greater than 1.0, interpreting as percentage...')
            m.State_of_charge_min = m.State_of_charge_min/100

        if not hasattr(m, 'State_of_charge_max'):
            # Fraction of energy capacity representing the maximum state of charge.
            logging.debug('ValuationOptimizer: No State_of_charge_max provided, setting default...')
            m.State_of_charge_max = 1
        elif getattr(m, 'State_of_charge_max') > 1.0:
            logging.warning('ValuationOptimizer: State_of_charge_max provided is greater than 1.0, interpreting as percentage...')
            m.State_of_charge_max = m.State_of_charge_max/100
        
        if not hasattr(m, 'State_of_charge_init'):
            # Initial state of charge as a fraction of energy capacity.
            logging.debug('ValuationOptimizer: No State_of_charge_init provided, setting default...')
            m.State_of_charge_init = 0.5
        elif getattr(m, 'State_of_charge_init') > 1.0:
            logging.warning('ValuationOptimizer: State_of_charge_init provided is greater than 1.0, interpreting as percentage...')
            m.State_of_charge_init = m.State_of_charge_init/100

        # Check if params necessary for certain market types are set if required.
        if self.market_type in {'ercot_arbreg', 'pjm_pfp', 'miso_pfp', 'isone_pfp', 'nyiso_pfp', 'spp_pfp', 'caiso_pfp'}:
            try:
                if not getattr(m, 'fraction_reg_up', None):
                    logging.debug('ValuationOptimizer: No fraction_reg_up provided, setting default...')
                    m.fraction_reg_up = 0.25
            except ValueError:  # fraction_reg_up is array-like
                if np.isnan(m.fraction_reg_up).any():
                    logging.debug('ValuationOptimizer: fraction_reg_up array-like provided has None values, setting default...')
                    m.fraction_reg_up = 0.25

            try:
                if not getattr(m, 'fraction_reg_down', None):
                    logging.debug('ValuationOptimizer: No fraction_reg_down provided, setting default...')
                    m.fraction_reg_down = 0.25
            except ValueError:  # fraction_reg_down is array-like
                if np.isnan(m.fraction_reg_down).any():
                    logging.debug('ValuationOptimizer: fraction_reg_down array-like provided has None values, setting default...')
                    m.fraction_reg_down = 0.25

            # Converts fraction_reg_up and fraction_reg_down to arrays.
            try:
                m.fraction_reg_up[len(m.price_electricity) - 1]
            except TypeError:
                m.fraction_reg_up = np.array([m.fraction_reg_up] * len(m.price_electricity))

            try:
                m.fraction_reg_down[len(m.price_electricity) - 1]
            except TypeError:
                m.fraction_reg_down = np.array([m.fraction_reg_down] * len(m.price_electricity))

        ###############################################################################################################
        if self.market_type in {'pjm_pfp', 'miso_pfp', 'nyiso_pfp', 'isone_pfp'}: # TODO: Figure out ISO NE for this?
            # Performance score for regulation service for pay-for-performance models.
            try:
                if not getattr(m, 'perf_score', None):
                    logging.debug('ValuationOptimizer: No perf_score provided, setting default...')
                    m.perf_score = 0.95
            except ValueError:  # perf_score is array-like
                if np.isnan(m.perf_score).any():
                    logging.debug('ValuationOptimizer: perf_score array-like provided has None values, setting default...')
                    m.perf_score = 0.95

            # Converts perf_score to array.
            try:
                # TODO: what if it is an array but longer than price of electricity??
                m.perf_score[len(m.price_electricity) - 1]
            except TypeError:
                m.perf_score = np.array([m.perf_score] * len(m.price_electricity))
            except IndexError:
                logging.warning('ValuationOptimizer: A perf_score array was provided but is shorter than the price_electricity array.')
                raise(IncompatibleDataException('ValuationOptimizer: There was a mismatch in array sizes between perf_score and price_electricity.'))
        
        if self.market_type in {'caiso_pfp'}: # TODO: Figure out SPP for this?
            # Performance score for regulation up and down services for pay-for-performance models.
            try:
                if not getattr(m, 'perf_score_ru', None):
                    logging.debug('ValuationOptimizer: No perf_score_ru provided, setting default...')
                    m.perf_score_ru = 0.95
            except ValueError:  # fraction_reg_up is array-like
                if np.isnan(m.perf_score_ru).any():
                    logging.debug('ValuationOptimizer: perf_score_ru array-like provided has None values, setting default...')
                    m.perf_score_ru = 0.95

            # Performance score for regulation up and down services for pay-for-performance models.
            try:
                if not getattr(m, 'perf_score_rd', None):
                    logging.debug('ValuationOptimizer: No perf_score_rd provided, setting default...')
                    m.perf_score_rd = 0.95
            except ValueError:  # fraction_reg_up is array-like
                if np.isnan(m.perf_score_rd).any():
                    logging.debug('ValuationOptimizer: perf_score_rd array-like provided has None values, setting default...')
                    m.perf_score_rd = 0.95

            # Converts perf_score_ru and perf_score_rd to arrays.
            try:
                # TODO: what if it is an array but longer than price of electricity, better to prompt an error??
                m.perf_score_ru[len(m.price_electricity) - 1]
            except TypeError:
                m.perf_score_ru = np.array([m.perf_score_ru] * len(m.price_electricity))
            except IndexError:
                logging.warning('ValuationOptimizer: A perf_score_ru array was provided but is shorter than the price_electricity array.')
                raise(IncompatibleDataException('ValuationOptimizer: There was a mismatch in array sizes between perf_score_ru and price_electricity.'))

            try:
                # TODO: what if it is an array but longer than price of electricity, better to prompt an error??
                m.perf_score_rd[len(m.price_electricity) - 1]
            except TypeError:
                m.perf_score_rd = np.array([m.perf_score_rd] * len(m.price_electricity))
            except IndexError:
                logging.warning('ValuationOptimizer: A perf_score_rd array was provided but is shorter than the price_electricity array.')
                raise(IncompatibleDataException('ValuationOptimizer: There was a mismatch in array sizes between perf_score_rd and price_electricity.'))

        if self.market_type in {'miso_pfp'}:
            if not hasattr(m, 'Make_whole'):
                # Adjusts the credit for providing regulation services in MISO.
                m.Make_whole = 0.03

    def _set_model_var(self):
        """Sets the model vars for the Pyomo ConcreteModel."""
        m = self.model

        if not hasattr(m, 's'):
            def _s_init(_m, t):
                """The energy storage device's state of charge [MWh]."""
                return m.State_of_charge_init*m.Energy_capacity

            m.s = Var(m.soc_time, initialize=_s_init, within=NonNegativeReals)

        if not hasattr(m, 'q_r'):
            def _q_r_init(_m, t):
                """The quantity of energy allocated for charging via energy arbitrage [MWh]."""
                return 0.0

            m.q_r = Var(m.time, initialize=_q_r_init, within=NonNegativeReals)

        if not hasattr(m, 'q_d'):
            def _q_d_init(_m, t):
                """The quantity of energy allocated for discharging via energy arbitrage [MWh]."""
                return 0.0

            m.q_d = Var(m.time, initialize=_q_d_init, within=NonNegativeReals)

        if not hasattr(m, 'q_ru'):
            def _q_ru_init(_m, t):
                """The quantity of energy offered into the regulation up market [MWh]."""
                return 0.0

            m.q_ru = Var(m.time, initialize=_q_ru_init, within=NonNegativeReals)

        if not hasattr(m, 'q_rd'):
            def _q_rd_init(_m, t):
                """The quantity of energy offered into the regulation down market [MWh]."""
                return 0.0

            m.q_rd = Var(m.time, initialize=_q_rd_init, within=NonNegativeReals)

        if not hasattr(m, 'q_reg'):
            def _q_reg_init(_m, t):
                """The quantity of energy offered into the regulation market [MWh]. For single product frequency regulation markets."""
                return 0.0

            m.q_reg = Var(m.time, initialize=_q_reg_init, within=NonNegativeReals)

    def instantiate_model(self):
        """Instantiates the Pyomo ConcreteModel and populates it with supplied time series data."""
        if not hasattr(self, 'model'):
            self.model = ConcreteModel()

        m = self.model

        try:
            m.time = RangeSet(0, len(self.price_electricity) - 1)
            m.soc_time = RangeSet(0, len(self.price_electricity))
        except TypeError:
            # self.price_electricity is of type 'NoneType'
            m.time = []
            m.soc_time = []

        m.price_electricity = self.price_electricity

        m.price_regulation = self.price_regulation
        m.price_reg_up = self.price_reg_up
        m.price_reg_down = self.price_reg_down

        m.price_reg_service = self.price_reg_service
        m.price_reg_serv_up = self.price_reg_serv_up
        m.price_reg_serv_down = self.price_reg_serv_down

        m.cost_charge = self.cost_charge
        m.cost_discharge = self.cost_discharge

        # m.reg_a = self.mileage_slow
        # m.reg_d = self.mileage_fast
        m.mi_mult = self.mileage_mult
        m.mi_mult_ru = self.mileage_mult_ru
        m.mi_mult_rd = self.mileage_mult_rd

        # If fraction_reg_up/fraction_reg_down are provided to the instance, set them in the ConcreteModel.
        if self.fraction_reg_up is not None:
            m.fraction_reg_up = self.fraction_reg_up
        if self.fraction_reg_down is not None:
            m.fraction_reg_down = self.fraction_reg_down

        if self.perf_score is not None:
            m.perf_score = self.perf_score
        if self.perf_score_ru is not None:
            m.perf_score_ru = self.perf_score_ru
        if self.perf_score_rd is not None:
            m.perf_score_rd = self.perf_score_rd

    def populate_model(self):
        """Populates the Pyomo ConcreteModel based on the specified market_type."""
        self.model.objective_expr = 0.0

        self._set_model_param()
        self._set_model_var()

        self.expressions_block = ExpressionsBlock(self.market_type)

        try:
            self.expressions_block.set_expressions(self.model)
        except IndexError:
            # Array-like object(s) do(es) not match the length of the price_electricity array-like.
            raise(IncompatibleDataException('At least one of the array-like parameter objects is not the expected length. (It should match the length of the price_electricity object.)'))
        else:
            self.model.objective = Objective(expr=self.model.objective_expr, sense=maximize)
        
        # if self.model.objective.value == 0.0:
        #     # Detect constant objective function value.
        #     raise(IncompatibleDataException('The objective function was ill-formed, resulting in a constant objective function.'))


    def _process_results(self):
        """Processes optimization results for further evaluation."""
        m = self.model

        t = m.time
        q_r = [m.q_r[n].value for n in m.time]
        q_d = [m.q_d[n].value for n in m.time]
        q_ru = [m.q_ru[n].value for n in m.time]
        q_rd = [m.q_rd[n].value for n in m.time]
        q_reg = [m.q_reg[n].value for n in m.time]
        soc = [m.s[n].value for n in m.time]
        price_electricity = [m.price_electricity[n] for n in m.time]

        run_results = {'time': t, 'q_r': q_r, 'q_d': q_d, 'q_ru': q_ru, 'q_rd': q_rd, 'q_reg': q_reg,
                       'state of charge': soc, 'price of electricity': price_electricity}

        if self.market_type == 'pjm_pfp':
            rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value) for t in m.time]))
            rev_reg = np.cumsum(np.array([m.q_reg[t].value*m.perf_score[t]*(m.mi_mult[t]*m.price_reg_service[t] + m.price_regulation[t]) for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue
        elif self.market_type == 'miso_pfp':
            rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value) for t in m.time]))
            rev_reg = np.cumsum(np.array([(1 + m.Make_whole)*m.perf_score[t]*m.price_regulation[t]*m.q_reg[t].value for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue
        #//////////////////////////////////////////////////#
        elif self.market_type == 'isone_pfp':
            rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value) for t in m.time]))
            rev_reg = np.cumsum(np.array([(m.mi_mult[t] * m.price_reg_service[t] + m.price_regulation[t]) * m.perf_score[t]*m.q_reg[t].value for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue
        #//////////////////////////////////////////////////#
        #######################################################################################################################
        elif self.market_type == 'nyiso_pfp':
            rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value) for t in m.time]))
            rev_reg = np.cumsum(np.array([m.q_reg[t].value* m.price_regulation[t] * (1 - 1.1*(1 - m.perf_score[t]))
                                          + m.price_electricity[t]*(m.q_reg[t].value * m.fraction_reg_up[t]
                                          - m.q_reg[t].value * m.fraction_reg_down[t])
                                          for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue
        #######################################################################################################################
        #######################################################################################################################
        elif self.market_type == 'spp_pfp':
            # TODO: copied from 'ercot_arbreg' -make sure is correct for this market
            # rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value
            #                                                       + m.q_ru[t].value*m.fraction_reg_up[t]
            #                                                       - m.q_rd[t].value*m.fraction_reg_down[t])
            #                               for t in m.time]))
            # rev_reg = np.cumsum(np.array([m.ru[t]*m.q_ru[t].value + m.rd[t]*m.q_rd[t].value for t in m.time]))
            rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value) for t in m.time]))
            rev_reg = np.cumsum(np.array([m.price_reg_up[t] * m.q_ru[t].value + m.price_reg_down[t] * m.q_rd[t].value
                                          + m.price_electricity[t]*(m.q_ru[t].value * m.fraction_reg_up[t]
                                          - m.q_rd[t].value * m.fraction_reg_down[t])
                                          for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue
        #######################################################################################################################
        elif self.market_type == 'caiso_pfp':
            # rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value
            #                                                       + m.q_ru[t].value*m.fraction_reg_up[t]
            #                                                       - m.q_rd[t].value*m.fraction_reg_down[t])
            #                               for t in m.time]))
            # rev_reg = np.cumsum(np.array([m.price_reg_up[t] * m.q_ru[t].value + m.price_reg_down[t] * m.q_rd[t].value
            #                               + m.perf_score_ru[t] * m.mi_mult_ru[t] * m.price_reg_serv_up[t]
            #                               + m.perf_score_rd[t] * m.mi_mult_rd[t] * m.price_reg_serv_down[t]
            #                               for t in m.time]))
            rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value) for t in m.time]))
            rev_reg = np.cumsum(np.array([m.price_reg_up[t] * m.q_ru[t].value + m.price_reg_down[t] * m.q_rd[t].value
                                          + m.perf_score_ru[t] * m.mi_mult_ru[t] * m.price_reg_serv_up[t]
                                          + m.perf_score_rd[t] * m.mi_mult_rd[t] * m.price_reg_serv_down[t]
                                          + m.price_electricity[t]*(m.q_ru[t].value * m.fraction_reg_up[t]
                                          - m.q_rd[t].value * m.fraction_reg_down[t])
                                          for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue
        #######################################################################################################################
        elif self.market_type == 'ercot_arbreg':
            # rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value
            #                                                       + m.q_ru[t].value*m.fraction_reg_up[t]
            #                                                       - m.q_rd[t].value*m.fraction_reg_down[t])
            #                               for t in m.time]))
            # # rev_reg = np.cumsum(np.array([m.ru[t] * m.q_ru[t].value + m.rd[t] * m.q_rd[t].value for t in m.time]))
            # rev_reg = np.cumsum(np.array([m.price_reg_up[t] * m.q_ru[t].value + m.price_reg_down[t] * m.q_rd[t].value for t in m.time]))
            rev_arb = np.cumsum(np.array([m.price_electricity[t]*(m.q_d[t].value - m.q_r[t].value) for t in m.time]))
            rev_reg = np.cumsum(np.array([m.price_reg_up[t] * m.q_ru[t].value + m.price_reg_down[t] * m.q_rd[t].value
                                          + m.price_electricity[t]*(m.q_ru[t].value * m.fraction_reg_up[t]
                                          - m.q_rd[t].value * m.fraction_reg_down[t])
                                          for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue
        else:
            rev_arb = np.cumsum(np.array([m.price_electricity[t] * (m.q_d[t].value - m.q_r[t].value)
                                          for t in m.time]))
            rev_reg = np.cumsum(np.array([0 for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results['rev_arb'] = rev_arb
            run_results['rev_reg'] = rev_reg
            run_results['revenue'] = revenue

        try:
            self.gross_revenue = revenue[-1]
        except IndexError:
            # Revenue is of length-0, likely due to no price_electricity array-like being given before solving.
            self.gross_revenue = 0
        
        self.results = pd.DataFrame(run_results)

    def get_results(self):
        """Returns the decision variables and derived quantities in a DataFrame, plus the net revenue."""
        return self.results, self.gross_revenue


class BadParameterException(Exception):
    pass


class IncompatibleDataException(Exception):
    pass


if __name__ == '__main__':
    with open('valuation_optimizer.log', 'w'):
        pass

    logging.basicConfig(filename='valuation_optimizer.log', format='[%(levelname)s] %(asctime)s: %(message)s',
                        level=logging.INFO)


from __future__ import division, print_function, absolute_import

import logging

from pyomo.environ import *
import pandas as pd
import numpy as np
import os

from es_gui.tools import optimizer
from es_gui.tools.btm.constraints import ExpressionsBlock


class BtmOptimizer(optimizer.Optimizer):
    """A framework wrapper class for creating Pyomo ConcreteModels for behind the meter valuation."""

    def __init__(self, tou_energy_schedule = None, tou_energy_rate=None, 
                 tou_demand_schedule=None, tou_demand_rate=None, flat_demand_rate=None,
                 nem_type=1, nem_rate=None, load_profile=None, pv_profile=None, 
                 rate_structure_metadata=None, load_profile_metadata=None, pv_profile_metadata=None,
                 cost_charge=None, cost_discharge=None,
                 solver='glpk'):
        
        self._model = ConcreteModel()
        self._solver = solver

        self._expressions_block = None
        
        self._tou_energy_schedule = tou_energy_schedule # type: list, size: number of hours in a month, value: tou_energy_rate index 
        self._tou_energy_rate = tou_energy_rate # type = list, size = number of tou periods for energy, value: tou energy rate [$/kWh] 
        
        self._tou_demand_schedule = tou_demand_schedule # type: list, size: number of hours in a month, value: tou_demand_rate index
        self._tou_demand_rate = tou_demand_rate # type = list, size = number of tou periods for demand , value:tou demand rate [$/kW]
        self._flat_demand_rate = flat_demand_rate # type = list, size = 12 (months) 
        
        self._nem_type = nem_type # type: integer, value: 0 = no net-metering, 1=net-metering 1.0, 2 = net-metering 2.0
        self._nem_rate = nem_rate # type: real, value: sell rate of nem 1.0
        
        self._load_profile = load_profile # type: list, size: number of hours in a month, value: load (kW) 
        self._pv_profile = pv_profile # type: list, size: number of hours in a month, value: pv (kW)

        self._rate_structure_metadata = rate_structure_metadata # type: dict 
        self._load_profile_metadata = load_profile_metadata # type: dict 
        self._pv_profile_metadata = pv_profile_metadata # type: dict
        
#        self._cost_charge = cost_charge
#        self._cost_discharge = cost_discharge
        self._results = None

        self._total_bill_with_es = 0
        self._total_bill_without_es = 0
        self._energy_charge_with_es = 0
        self._energy_charge_without_es = 0
        self._demand_charge_with_es = 0
        self._demand_charge_without_es = 0
        self._nem_charge_with_es = 0
        self._nem_charge_without_es = 0
        
 #---------------------------------------------------   
    @property
    def tou_energy_schedule(self):
        """Time-of-use energy schedule."""
        return self._tou_energy_schedule

    @tou_energy_schedule.setter
    def tou_energy_schedule(self, value):
        self._tou_energy_schedule = value
 #--------------------------------------------------- 
    @property
    def tou_energy_rate(self):
        """Time-of-use energy rate [$/kWh]."""
        return self._tou_energy_rate

    @tou_energy_rate.setter
    def tou_energy_rate(self, value):
        self._tou_energy_rate = value
 #--------------------------------------------------- 
    @property
    def tou_demand_schedule(self):
        """Time-of-use demand schedule."""
        return self._tou_demand_schedule

    @tou_demand_schedule.setter
    def tou_demand_schedule(self, value):
        self._tou_demand_schedule = value
 #---------------------------------------------------  
    @property
    def tou_demand_rate(self):
        """Time-of-use demand rate [$/kW]."""
        return self._tou_demand_rate

    @tou_demand_rate.setter
    def tou_demand_rate(self, value):
        self._tou_demand_rate = value
 #--------------------------------------------------- 
    @property
    def flat_demand_rate(self):
        """Flat demand rate [$/kW]."""
        return self._flat_demand_rate

    @flat_demand_rate.setter
    def flat_demand_rate(self, value):
        self._flat_demand_rate = value
#--------------------------------------------------- 
    @property
    def nem_type(self):
        """The type of net metering program, defaults to 1."""
        return self._nem_type

    @nem_type.setter
    def nem_type(self, value):
        if isinstance(value, int):
            self._nem_type = value
        else:
            raise TypeError('nem_type property must be of type int.')
 #--------------------------------------------------- 
    @property
    def nem_rate(self):
        """The type of net metering program"""
        return self._nem_rate

    @nem_rate.setter
    def nem_rate(self, value):
        self._nem_rate = value 
 #--------------------------------------------------- 
    @property
    def load_profile(self):
        """Load profile [kW]."""
        return self._load_profile

    @load_profile.setter
    def load_profile(self, value):
        self._load_profile = value
 #--------------------------------------------------- 
    @property
    def pv_profile(self):
        """Pv profile [kW]."""
        return self._pv_profile 

    @pv_profile.setter
    def pv_profile (self, value):
        self._pv_profile = value
#--------------------------------------------------- 
    @property
    def rate_structure_metadata(self):
        """Dictionary containing metadata about the rate structure associated with the other input properties."""
        return self._rate_structure_metadata 

    @rate_structure_metadata.setter
    def rate_structure_metadata (self, value):
        self._rate_structure_metadata = value
#--------------------------------------------------- 
    @property
    def load_profile_metadata(self):
        """Dictionary containing metadata about the load profile."""
        return self._load_profile_metadata 

    @load_profile_metadata.setter
    def load_profile_metadata (self, value):
        self._load_profile_metadata = value
#--------------------------------------------------- 
    @property
    def pv_profile_metadata(self):
        """Dictionary containing metadata about the PV profile."""
        return self._pv_profile_metadata 

    @pv_profile_metadata.setter
    def pv_profile_metadata (self, value):
        self._pv_profile_metadata = value
# #--------------------------------------------------- 
#    @property
#    def cost_charge(self):
#        """The cost of charging the energy storage device [$/kWh]."""
#        return self._cost_charge
#
#    @cost_charge.setter
#    def cost_charge(self, value):
#        self._cost_charge = value
# #--------------------------------------------------- 
#    @property
#    def cost_discharge(self):
#        """The cost of discharging the energy storage device [$/kWh]."""
#        return self._cost_discharge
#
#    @cost_discharge.setter
#    def cost_discharge(self, value):
#        self._cost_discharge = value
 #--------------------------------------------------- 
    @property
    def solver(self):
        """The name of the solver for Pyomo to use, defaults to 'glpk'."""
        return self._solver

    @solver.setter
    def solver(self, value):
        self._solver = value
 #--------------------------------------------------- 
    @property
    def expressions_block(self):
        """ExpressionsBlock object for setting model objectives and constraints."""
        return self._expressions_block

    @expressions_block.setter
    def expressions_block(self, value):
        self._expressions_block = value
 
 #--------------------------------------------------- 
    @property
    def results(self):
        """Pandas DataFrame containing results."""
        return self._results

    @results.setter
    def results(self, value):
        if isinstance(value, pd.DataFrame):
            self._results = value
        else:
            raise TypeError('results must be a Pandas DataFrame.')
 #--------------------------------------------------- 
    @property
    def total_bill_with_es(self):
        """The total bill for the month with energy storage."""
        return self._total_bill_with_es

    @total_bill_with_es.setter
    def total_bill_with_es(self, value):
        self._total_bill_with_es = value
    
    @property
    def total_bill_without_es(self):
        """The total bill for the month without energy storage."""
        return self._total_bill_without_es

    @total_bill_without_es.setter
    def total_bill_without_es(self, value):
        self._total_bill_without_es = value
    
    @property
    def demand_charge_with_es(self):
        """The total demand charges for the month with energy storage."""
        return self._demand_charge_with_es

    @demand_charge_with_es.setter
    def demand_charge_with_es(self, value):
        self._demand_charge_with_es = value
    
    @property
    def demand_charge_without_es(self):
        """The total demand charges for the month without energy storage."""
        return self._demand_charge_without_es

    @demand_charge_without_es.setter
    def demand_charge_without_es(self, value):
        self._demand_charge_without_es = value
    
    @property
    def energy_charge_with_es(self):
        """The total energy charges for the month with energy storage."""
        return self._energy_charge_with_es

    @energy_charge_with_es.setter
    def energy_charge_with_es(self, value):
        self._energy_charge_with_es = value
    
    @property
    def energy_charge_without_es(self):
        """The total energy charges for the month without energy storage."""
        return self._energy_charge_without_es

    @energy_charge_without_es.setter
    def energy_charge_without_es(self, value):
        self._energy_charge_without_es = value
    
    @property
    def nem_charge_with_es(self):
        """The total net energy metering charges for the month with energy storage."""
        return self._nem_charge_with_es

    @nem_charge_with_es.setter
    def nem_charge_with_es(self, value):
        self._nem_charge_with_es = value
    
    @property
    def nem_charge_without_es(self):
        """The total net energy metering charges for the month without energy storage."""
        return self._nem_charge_without_es

    @nem_charge_without_es.setter
    def nem_charge_without_es(self, value):
        self._nem_charge_without_es = value

    def _set_model_param(self):
        """Sets the model params for the Pyomo ConcreteModel."""
        m = self.model
        
        # Check if params common to all formulations are set.
        if not hasattr(m, 'Transfomer_rating'):
            # Transformer rating; equivalently, the maximum power can be exchanged [kW].
            logging.debug('Optimizer: No Transformer_rating provided, setting default...')
            m.Transformer_rating = 1000000
            
        if not hasattr(m, 'Power_rating'):
            # Power rating; equivalently, the maximum power can be charged or discharged [kW].
            logging.debug('Optimizer: No Power_rating provided, setting default...')
            m.Power_rating = 100

        if not hasattr(m, 'Energy_capacity'):
            # Energy capacity [kWh].
            logging.debug('Optimizer: No Energy_capacity provided, setting default...')
            m.Energy_capacity = 100

        if not hasattr(m, 'Self_discharge_efficiency'):
            # Fraction of energy maintained over one time period.
            logging.debug('Optimizer: No Self_discharge_efficiency provided, setting default...')
            m.Self_discharge_efficiency = 1.00     
        elif getattr(m, 'Self_discharge_efficiency') > 1.0:
            logging.warning('Optimizer: Self_discharge_efficiency provided is greater than 1.0, interpreting as percentage...')
            m.Self_discharge_efficiency = m.Self_discharge_efficiency/100

        if not hasattr(m, 'Round_trip_efficiency'):
            # Fraction of input energy that gets stored over one time period.
            logging.debug('Optimizer: No Round_trip_efficiency provided, setting default...')
            m.Round_trip_efficiency = 0.85
        elif getattr(m, 'Round_trip_efficiency') > 1.0:
            logging.warning('Optimizer: Round_trip_efficiency provided is greater than 1.0, interpreting as percentage...')
            m.Round_trip_efficiency = m.Round_trip_efficiency/100

        if not hasattr(m, 'State_of_charge_min'):
            # Fraction of energy capacity to increase state of charge minimum by.
            logging.debug('Optimizer: No State_of_charge_min provided, setting default...')
            m.State_of_charge_min = 0
        elif getattr(m, 'State_of_charge_min') > 1.0:
            logging.warning('Optimizer: State_of_charge_min provided is greater than 1.0, interpreting as percentage...')
            m.State_of_charge_min = m.State_of_charge_min/100

        if not hasattr(m, 'State_of_charge_max'):
            # Fraction of energy capacity to decrease state of charge maximum by.
            logging.debug('Optimizer: No State_of_charge_max provided, setting default...')
            m.State_of_charge_max = 100
        elif getattr(m, 'State_of_charge_max') > 1.0:
            logging.warning('Optimizer: State_of_charge_max provided is greater than 1.0, interpreting as percentage...')
            m.State_of_charge_max = m.State_of_charge_max/100
        
        if not hasattr(m, 'State_of_charge_init'):
            # Initial state of charge [fraction of capacity], defaults to the amount reserved for discharging.
            logging.debug('Optimizer: No State_of_charge_init provided, setting default...')
            m.State_of_charge_init = 0.50
        elif getattr(m, 'State_of_charge_init') > 1.0:
            logging.warning('Optimizer: State_of_charge_init provided is greater than 1.0, interpreting as percentage...')
            m.State_of_charge_init = m.State_of_charge_init/100
            
        m.smin = m.State_of_charge_min*m.Energy_capacity
        m.smax = m.State_of_charge_max*m.Energy_capacity
    
    def _set_model_var(self):
        """Sets the model vars for the Pyomo ConcreteModel."""
        m = self.model
        
        if not hasattr(m, 's'):
            def _s_init(_m, t):
                """The energy storage device's state of charge [kWh]."""
                return m.State_of_charge_init*m.Energy_capacity

            m.s = Var(m.time, domain = NonNegativeReals,bounds=(m.smin,m.smax))

        if not hasattr(m, 'pdis'):
            def _pdis_init(_m, t):
                """The discharge power vector [kW]"""
                return 0.0

            m.pdis = Var(m.time, initialize= _pdis_init, domain = NonNegativeReals, bounds=(0,m.Power_rating))
        
        if not hasattr(m, 'pcha'):
            def _pcha_init(_m, t):
                """The charge power vector [kW]"""
                return 0.0

            m.pcha = Var(m.time, initialize= _pcha_init, domain = NonNegativeReals, bounds=(0,m.Power_rating))
        
        if not hasattr(m, 'pfpk'):
            def _pfpk_init(_m):
                """Peak demand of a month [kW]"""
                return 0.0

            m.pfpk = Var(initialize= _pfpk_init, domain = NonNegativeReals, bounds=(0,m.Transformer_rating))
            
        if not hasattr(m, 'ptpk'):
            def _ptpk_init(_m, p):
                """Peak demand of different time-of-use periods of a month [kW]"""
                return 0.0

            m.ptpk = Var(m.period, initialize= _ptpk_init, domain = NonNegativeReals,bounds=(0,m.Transformer_rating))
        
        if not hasattr(m, 'xnet'):
            def _xnet_init(_m, t):
                """Xnet[i]=Max{Pload[i]-Ppv[i]+Pcha[i]-Pdis[i],0} [kW]"""
                return 0.0

            m.xnet = Var(m.time, initialize= _xnet_init, domain = NonNegativeReals,bounds=(0,m.Transformer_rating))
        
    def instantiate_model(self):
        """Instantiates the Pyomo ConcreteModel and populates it with supplied time series data."""
        if not hasattr(self, 'model'):
            self.model = ConcreteModel()

        m = self.model
        m.nhr = len(self.tou_energy_schedule)
        m.dml = len(self.tou_demand_rate)
        
        try:
            m.time = RangeSet(0, m.nhr - 1)
            
        except TypeError:
            # self.tou_energy_schedule is of type 'NoneType'
            m.time = []
        
        try:
            m.period = RangeSet(0, m.dml - 1)
            
        except TypeError:
            # self.tou_demand_rate is of type 'NoneType'
            m.period = []
        
        m.tou_er = [self.tou_energy_rate[self.tou_energy_schedule[t]] for t in range(m.nhr)]
        
        m.tou_dr = self.tou_demand_rate
        
        mask_ds = []
        for i in range(m.dml):
            listi=[int(self.tou_demand_schedule[t]==i) for t in range(m.nhr)]
            mask_ds.append(listi)
       
        m.mask_ds = mask_ds
        
        m.flt_dr = self.flat_demand_rate
        
        if self.nem_type==0:
            m.nem_sr=[0 for t in range(m.nhr)]
        elif self.nem_type==1:
            m.nem_sr=[self.nem_rate for t in range(m.nhr)]
        else:
            m.nem_sr=m.tou_er
        
        m.pld = self.load_profile
        m.ppv = self.pv_profile
        m.pnet= [m.pld[t]-m.ppv[t] for t in range(m.nhr)]
        
#        m.cost_cha = self.cost_charge
#        m.cost_dis = self.cost_discharge

              
    def populate_model(self):
        """Populates the Pyomo ConcreteModel based on the specified market_type."""
        self.model.objective_expr = 0.0

        self._set_model_param()
        self._set_model_var()

        self.expressions_block = ExpressionsBlock()

        try:
            self.expressions_block.set_expressions(self.model)
        except IndexError:
            # Array-like object(s) do(es) not match the length of the price_electricity array-like.
            raise(IncompatibleDataException('At least one of the array-like parameter objects is not the expected length. (It should match the length of the price_electricity object.)'))
        else:
            self.model.objective = Objective(expr=self.model.objective_expr)

    def _process_results(self):
        """Processes optimization results for further evaluation."""
        m = self.model

        pcha = [m.pcha[n].value for n in m.time]
        pdis = [m.pdis[n].value for n in m.time]
        ptot = [m.pnet[n] + m.pcha[n].value - m.pdis[n].value for n in m.time]
        soc  = [m.s[n].value for n in m.time]
        pfpk_without_es = max(m.pnet)
        ptpk_without_es = [max(m.pnet[n]*m.mask_ds[p][n] for n in m.time) for p in m.period]
        
        demand_charge_with_es=m.pfpk.value*m.flt_dr+sum(m.ptpk[p].value*m.tou_dr[p] for p in m.period)
        demand_charge_without_es=pfpk_without_es*m.flt_dr+sum(ptpk_without_es[p]*m.tou_dr[p] for p in m.period)
        
        energy_charge_with_es=sum(max(0,ptot[n])*m.tou_er[n] for n in m.time)
        energy_charge_without_es=sum(max(0,m.pnet[n])*m.tou_er[n] for n in m.time)
        
        nem_charge_with_es=sum(min(0,ptot[n])*m.nem_sr[n] for n in m.time) #negative since it is credit
        nem_charge_without_es=sum(min(0,m.pnet[n])*m.nem_sr[n] for n in m.time) #negative since it is credit
        
        tot_bill_with_es=demand_charge_with_es + energy_charge_with_es + nem_charge_with_es
        tot_bill_without_es=demand_charge_without_es + energy_charge_without_es + nem_charge_without_es
        
        run_results = {'time': m.time, 'Pload': m.pld, 'Ppv': m.ppv, 'Pcharge': pcha, 'Pdischarge': pdis, 'Ptotal': ptot,
                       'state of charge': soc, 'energy_charge_with_es': energy_charge_with_es,'nem_charge_with_es': nem_charge_with_es, 
                       'demand_charge_with_es':demand_charge_with_es, 'total_bill_with_es': tot_bill_with_es, 
                       'energy_charge_without_es': energy_charge_without_es,'nem_charge_without_es': nem_charge_without_es, 
                       'demand_charge_without_es':demand_charge_without_es, 'total_bill_without_es': tot_bill_without_es }
        self.results = pd.DataFrame(run_results)

        self.total_bill_with_es = tot_bill_with_es
        self.total_bill_without_es = tot_bill_without_es

        self.demand_charge_with_es = demand_charge_with_es
        self.demand_charge_without_es = demand_charge_without_es

        self.energy_charge_with_es = energy_charge_with_es
        self.energy_charge_without_es = energy_charge_without_es

        self.nem_charge_with_es = nem_charge_with_es
        self.nem_charge_without_es = nem_charge_without_es
        
    def get_results(self):
        """Returns the decision variables and derived quantities in a DataFrame"""
        return self.results
    
    def has_energy_charges(self):
        """Returns True if there are energy charges (savings)."""
        if abs(self.energy_charge_with_es - self.energy_charge_without_es) > 1e-4:
            return True
        else:
            return False
    
    def has_demand_charges(self):
        """Returns True if there are demand charges (savings)."""
        if abs(self.demand_charge_with_es - self.demand_charge_without_es) > 1e-4:
            return True
        else:
            return False
    
    def has_nem_charges(self):
        """Returns True if there are net metering charges (savings)."""
        if abs(self.nem_charge_with_es - self.nem_charge_without_es) > 1e-4:
            return True
        else:
            return False


class BadParameterException(Exception):
    pass


class IncompatibleDataException(Exception):
    pass


if __name__ == '__main__':
    with open('btm_optimizer.log', 'w'):
        pass

    logging.basicConfig(filename='btm_optimizer.log', format='[%(levelname)s] %(asctime)s: %(message)s',
                        level=logging.INFO)


# This is for setting the window parameters like the initial size. Goes before any other import statements.
from kivy.config import Config

Config.set('graphics', 'height', '900')
Config.set('graphics', 'width', '1600')
Config.set('graphics', 'minimum_height', '720')
Config.set('graphics', 'minimum_width', '1280')
Config.set('graphics', 'resizable', '1')
Config.set('kivy', 'desktop', 1)

from numpy.random import random, randn
import numpy as np

from kivy.utils import get_color_from_hex
from kivy.lang import Builder
from kivy.app import App
from kivy.uix.screenmanager import ScreenManager, Screen
from kivy.uix.boxlayout import BoxLayout
from kivy.core.text import LabelBase

from es_gui.proving_grounds import charts
import calendar


class ChartTestApp(App):
    """
    The App class for launching the application.
    """

    def build(self):
        # Sets the window/application title
        self.title = 'Charts'

        # Create ScreenManager
        sm = ScreenManager()
        sm.add_widget(ChartScreen())

        # Create BoxLayout container
        bx = BoxLayout(orientation='vertical')

        # Fill BoxLayout
        bx.add_widget(sm)

        return bx


class ChartScreen(Screen):
    def draw_chart(self, chart):
        bar_data = [
            ['Jan', random(4), randn()],
            ['Feb', random(4), randn()],
            ['Mar', random(4), randn()],
            ['Apr', random(4), randn()],
            ['May', random(4), randn()],
            ['Jun', random(4), randn()],
            ['Jul', random(4), randn()],
            ['Aug', random(4), randn()],
            ['Sep', random(4), randn()],
            ['Oct', random(4), randn()],
            ['Nov', random(4), randn()],
            ['Dec', random(4), randn()],
        ]
        chart.draw_chart(bar_data)

    def draw_stacked_chart(self, chart):
        from collections import OrderedDict

        rgba1 = random(4)
        rgba2 = random(4)
        rgba3 = random(4)

        bar_data = OrderedDict()
        bar_data['Jan'] = [['alpha', rgba1, random()], ['beta', rgba2, random()],
                           ['gamma', rgba3, random()]]
        bar_data['Feb'] = [['alpha', rgba1, random()], ['beta', rgba2, random()],
                           ['gamma', rgba3, random()]]
        bar_data['Mar'] = [['alpha', rgba1, random()], ['beta', rgba2, random()],
                           ['gamma', rgba3, random()]]
        bar_data['Apr'] = [['alpha', rgba1, random()], ['beta', rgba2, random()],
                           ['gamma', rgba3, random()]]
        bar_data['May'] = [['alpha', rgba1, random()], ['beta', rgba2, random()],
                           ['gamma', rgba3, random()]]

        chart.draw_chart(bar_data)

    def draw_multiset_chart(self, chart):
        from collections import OrderedDict

        rgba1 = random(4)
        rgba2 = random(4)
        rgba3 = random(4)

        bar_data = OrderedDict()
        bar_data['Jan'] = [['alpha', rgba1, randn()], ['beta', rgba2, randn()],
                           ['gamma', rgba3, randn()]]
        bar_data['Feb'] = [['alpha', rgba1, randn()], ['beta', rgba2, randn()],
                           ['gamma', rgba3, randn()]]
        bar_data['Mar'] = [['alpha', rgba1, randn()], ['beta', rgba2, randn()],
                           ['gamma', rgba3, randn()]]
        bar_data['Apr'] = [['alpha', rgba1, randn()], ['beta', rgba2, randn()],
                           ['gamma', rgba3, randn()]]
        bar_data['May'] = [['alpha', rgba1, randn()], ['beta', rgba2, randn()],
                           ['gamma', rgba3, randn()]]

        chart.draw_chart(bar_data)

    def draw_pie(self, chart, is_donut=False):
        from random import sample
        from string import ascii_letters
        sample(ascii_letters, 5)

        pie_data = [
            [''.join(sample(ascii_letters, 12)), random(4), random()]
            for x in range(5)
            ]

        chart.draw_chart(pie_data, is_donut=is_donut)

    def draw_donut(self, chart):
        from random import sample
        from string import ascii_letters

        pie_data = [
            [''.join(sample(ascii_letters, 12)), random(4), random()]
            for x in range(4)
        ]
        pie_data.append([''.join(sample(ascii_letters, 12)), random(4), random()])

        chart.draw_chart(pie_data)
    
    def draw_rate_schedule_chart(self, chart):
        n_tiers = np.random.randint(1, 4)
        # print(n_tiers)

        schedule_data = np.random.randint(n_tiers, size=(12,24))
        # print(schedule_data)

        labels = calendar.month_abbr[1:]
        category_colors = [random(4), random(4), random(4)]
        
        chart.draw_chart(schedule_data, category_colors[:n_tiers], labels)

if __name__ == '__main__':
    from kivy.core.window import Window

    # Sets window background color
    Window.clearcolor = get_color_from_hex('#FFFFFF')

    ChartTestApp().run()